from pathlib import Path

import genvarloader as gvl
from pytest_cases import fixture


@fixture
def pgen_variants():
    pgen_path = Path(__file__).parent / "data" / "pgen" / "filtered_sample.pgen"
    return gvl.Variants.from_file(pgen_path)


@fixture
def vcf_variants():
    vcf_path = Path(__file__).parent / "data" / "vcf" / "filtered_sample.vcf"
    return gvl.Variants.from_file(vcf_path)



import importlib.metadata

from loguru import logger

from ._bigwig import BigWigs
from ._dataset._impl import ArrayDataset, Dataset, RaggedDataset
from ._dataset._write import write
from ._dummy import get_dummy_dataset
from ._ragged import Ragged
from ._utils import read_bedlike, with_length
from ._variants import Variants

__version__ = importlib.metadata.version("genvarloader")

__all__ = [
    "write",
    "Dataset",
    "Variants",
    "BigWigs",
    "read_bedlike",
    "with_length",
    "Ragged",
    "get_dummy_dataset",
    "ArrayDataset",
    "RaggedDataset",
]


logger.disable("genvarloader")



from __future__ import annotations

from pathlib import Path
from typing import Dict, Optional

import cyvcf2
import numpy as np
import polars as pl
from attrs import define, field
from numpy.typing import ArrayLike, NDArray

from .._ragged import Ragged
from .._utils import _lengths_to_offsets, _normalize_contig_name
from ._genotypes import CCFFieldError, PgenGenos, VCFGenos
from ._records import Records
from ._utils import path_is_pgen, path_is_vcf

__all__ = []


@define
class DenseGenotypes:
    """Dense array(s) of genotypes."""

    var_start_idxs: NDArray[np.int32]
    """Shape: (regions)"""
    var_end_idxs: NDArray[np.int32]
    """Shape: (regions)"""
    genotypes: NDArray[np.int8]
    """Shape: (samples, ploid, variants). All (sample, ploid) have the same # of variants per region"""
    offsets: NDArray[np.int64]


@define
class DenseGenosAndCCF(DenseGenotypes):
    ccfs: NDArray[np.float32]
    """Shape: (samples, var: variants). Uses the same offsets as genotypes."""


@define
class Variants:
    """Variant records and genotypes. Should not be directly instantiated, use
    :meth:`from_file` instead."""

    records: Records
    genotypes: VCFGenos | PgenGenos
    phased: bool
    ccf_field: str | None = None
    _sample_idxs: NDArray[np.intp] | None = field(default=None, alias="_sample_idxs")

    @classmethod
    def from_file(
        cls,
        path: str | Path | Dict[str, Path],
        phased=True,
        ccf_field: str | None = None,
    ) -> "Variants":
        """Create a Variants instances from a VCF or PGEN file(s). If a dictionary is provided, the keys should be
        contig names and the values should be paths to the corresponding VCF or PGEN.

        Parameters
        ----------
        path
            Path to a VCF or PGEN file or a mapping from contig names to paths.
        phased
            Whether the genotypes should be treated as phased.
        ccf_field
            The name of the cancel cell fraction (CCF) field in the VCF file. This is currently only applicable and required for
            unphased genotypes.
        """
        if isinstance(path, (str, Path)):
            path = Path(path)
            first_path = path
        elif isinstance(path, dict):
            first_path = next(iter(path.values()))

        if path_is_vcf(first_path):
            return cls._from_vcf(path, phased, ccf_field)
        elif path_is_pgen(first_path):
            return cls._from_pgen(path, phased)
        else:
            raise ValueError("Unsupported file type.")

    @property
    def chunked(self) -> bool:
        return self.genotypes.chunked

    @property
    def samples(self) -> NDArray[np.str_]:
        if self._sample_idxs is None:
            return self.genotypes.samples
        return self.genotypes.samples[self._sample_idxs]

    @property
    def n_samples(self) -> int:
        if self._sample_idxs is None:
            return self.genotypes.n_samples
        return len(self._sample_idxs)

    @property
    def ploidy(self) -> int:
        if self.phased:
            return self.genotypes.ploidy
        else:
            return 1

    @classmethod
    def _from_vcf(
        cls,
        vcf: str | Path | Dict[str, Path],
        phased: bool,
        ccf_field: str | None,
    ) -> "Variants":
        """Currently does not support multi-allelic sites, but does support *split*
        multi-allelic sites. Note that SVs and "other" variants are also not supported.
        VCFs can be prepared by running:

        .. code-block:: bash

            bcftools view -i 'TYPE="snp" || TYPE="indel"' <file.bcf> \\
            | bcftools norm \\
            -a \\
            --atom-overlaps . \\
            -m - \\
            -f <ref.fa> \\
            -O b \\
            -o <norm.bcf>
        
        """
        if not phased and ccf_field is None:
            raise ValueError("CCF field is required for unphased genotypes.")

        elif not phased and ccf_field is not None:
            _vcf = cyvcf2.VCF(vcf)
            try:
                ccf_field_info = _vcf.get_header_type(ccf_field)
            except KeyError:
                raise CCFFieldError(
                    f"CCF field '{ccf_field}' not found in VCF header."
                )

            if ccf_field_info["Number"] not in {"1", "A"}:
                raise CCFFieldError(
                    f"CCF field '{ccf_field}' must have Number equal to '1' or 'A'."
                )

        records = Records.from_vcf(vcf)

        genotypes = VCFGenos(vcf, records.contig_offsets)
        return cls(records, genotypes, phased, ccf_field)

    @classmethod
    def _from_pgen(cls, pgen: str | Path | Dict[str, Path], phased: bool) -> "Variants":
        """Currently does not support multi-allelic sites, but does support *split*
        multi-allelic sites. Note that SVs and "other" variants are also not supported.
        A PGEN can be prepared from a VCF by running:

        .. code-block:: bash

            bcftools view -i 'TYPE="snp" || TYPE="indel"' <file.bcf> \\
            | bcftools norm \\
            -a \\
            --atom-overlaps . \\
            -m - \\
            -f <ref.fa> \\
            -O b \\
            -o <norm.bcf>

            plink2 --make-pgen \\
            --bcf <norm.bcf> \\
            --vcf-half-call r \\
            --out <prefix>
        
        """
        if isinstance(pgen, str):
            pgen = Path(pgen)

        if isinstance(pgen, Path):
            pgen = {"_all": pgen}

        psams: Dict[str, pl.DataFrame] = {}
        samples = None
        for contig, path in pgen.items():
            with open(path.with_suffix(".psam")) as f:
                cols = [c.strip("#") for c in f.readline().strip().split()]

            psam = pl.read_csv(
                path.with_suffix(".psam"),
                separator="\t",
                has_header=False,
                skip_rows=1,
                new_columns=cols,
                schema_overrides={
                    "FID": pl.Utf8,
                    "IID": pl.Utf8,
                    "SID": pl.Utf8,
                    "PAT": pl.Utf8,
                    "MAT": pl.Utf8,
                    "SEX": pl.Utf8,
                },
            )

            if samples is None:
                samples = psam["IID"].to_numpy()
            else:
                intersection = np.intersect1d(samples, psam["IID"].to_numpy())
                if np.any(samples != intersection):
                    raise ValueError("Sample names do not match across contigs.")

            psams[contig] = psam
        assert samples is not None

        records = Records.from_pvar(
            {c: p.with_suffix(".pvar") for c, p in pgen.items()}
        )

        genotypes = PgenGenos(pgen, samples, records.contigs)
        return cls(records, genotypes, phased)

    def _subset_samples(self, samples: ArrayLike):
        """Subset the samples to be read from the genotypes.

        Parameters
        ----------
        samples
            List of sample names."""
        samples = np.atleast_1d(np.asarray(samples, dtype=str))
        geno_sample_idxs, sample_idxs = np.intersect1d(
            self.genotypes.samples, samples, return_indices=True
        )[1:]
        if len(sample_idxs) != len(samples):
            raise ValueError("Some samples were not found")
        self._sample_idxs = geno_sample_idxs[sample_idxs]

    def _read(
        self,
        contig: str,
        starts: ArrayLike,
        ends: ArrayLike,
        sample: Optional[ArrayLike] = None,
        ploid: Optional[ArrayLike] = None,
    ) -> DenseGenotypes | tuple[DenseGenotypes, Ragged[np.float32]] | None:
        """Read genotypes for a given contig and range of positions.

        Parameters
        ----------
        contig
            Name of the contig/chromosome.
        starts
            Start coordinates, 0-based.
        ends
            End coordinates, 0-based exclusive.
        sample
            List of sample names.
        ploid
            List of ploid numbers.
        """
        _contig = _normalize_contig_name(contig, self.records.contigs)
        if _contig is None:
            return None

        starts = np.atleast_1d(np.asarray(starts, dtype=int))
        ends = np.atleast_1d(np.asarray(ends, dtype=int))

        if sample is not None:
            sample = np.atleast_1d(np.asarray(sample, dtype=str))
            _, geno_sample_idxs, sample_idxs = np.intersect1d(
                self.genotypes.samples, sample, return_indices=True
            )
            sample_idxs = geno_sample_idxs[sample_idxs]
            if len(sample_idxs) != len(sample):
                raise ValueError("Some samples were not found")
        else:
            sample_idxs = None

        if ploid is not None:
            ploid = np.atleast_1d(np.asarray(ploid, dtype=int))

        starts_ends = self.records.vars_in_range(_contig, starts, ends)
        if starts_ends is None:
            return None

        s_idxs, e_idxs = starts_ends
        n_variants = e_idxs - s_idxs

        ccfs = None
        # (s p v)
        if isinstance(self.genotypes, PgenGenos):
            if self.ccf_field is not None:
                raise NotImplementedError("Cannot read dosage from PGEN files.")

            # (s p v)
            genos = self.genotypes.read(_contig, s_idxs, e_idxs, sample_idxs, ploid)
        else:
            if self.ccf_field is None:
                genos = self.genotypes.read(
                    _contig, starts, ends, sample_idxs, ploid, n_variants
                )
            else:
                genos, ccfs = self.genotypes.read_genos_and_ccfs(
                    contig,
                    starts,
                    ends,
                    self.ccf_field,
                    sample_idxs,
                    ploid,
                    n_variants,
                )

        offsets = _lengths_to_offsets(n_variants)

        if ccfs is None:
            return DenseGenotypes(
                s_idxs,
                e_idxs,
                genos,
                offsets,
            )
        else:
            return DenseGenosAndCCF(
                s_idxs,
                e_idxs,
                genos,
                offsets,
                ccfs,
            )



from itertools import accumulate, chain, repeat
from pathlib import Path
from typing import (
    Any,
    Dict,
    Generator,
    Iterable,
    Optional,
    Sequence,
    TypeGuard,
    TypeVar,
    Union,
    cast,
)

import numpy as np
import pandera as pa
import pandera.typing as pat
import polars as pl
from numpy.typing import NDArray

from ._types import DTYPE, Idx

__all__ = [
    "read_bedlike",
    "with_length",
]

T = TypeVar("T")


def is_dtype(arr: NDArray, dtype: type[DTYPE]) -> TypeGuard[NDArray[DTYPE]]:
    return np.issubdtype(arr.dtype, dtype)


def _process_bed(bed: Union[str, Path, pl.DataFrame], fixed_length: int):
    if isinstance(bed, (str, Path)):
        bed = read_bedlike(bed)

    if "strand" in bed and bed["strand"].dtype == pl.Utf8:
        bed = bed.with_columns(
            pl.col("strand").replace({"-": -1, "+": 1}, return_dtype=pl.Int8)
        )
    else:
        bed = bed.with_columns(strand=pl.lit(1, dtype=pl.Int8))

    if "region_idx" not in bed:
        bed = bed.with_row_count("region_idx")

    bed = bed.sort("chrom", "chromStart")

    return with_length(bed, fixed_length)


def with_length(bed: pl.DataFrame, length: int) -> pl.DataFrame:
    """Expands or shrinks each region in a BED-like DataFrame to be fixed-length windows
    centered around the midpoint of each region or the "peak" column if it is present.

    .. important::

        The "peak" column is described in the `narrowPeak <https://genome.ucsc.edu/FAQ/FAQformat.html#format12>`_
        and `broadPeak <https://genome.ucsc.edu/FAQ/FAQformat.html#format13>`_ specifications. It is a 0-based
        offset from chromStart, so be sure not to encode your "peak" column as an absolute position!

    Parameters
    ----------
    bed
        BED-like DataFrame with at least the columns "chrom", "chromStart", and "chromEnd".
    length
        Length of the fixed-length windows.
    """
    if "peak" in bed:
        center = pl.col("chromStart") + pl.col("peak")
    else:
        center = (pl.col("chromStart") + pl.col("chromEnd")) // 2
    left = length // 2
    right = length - left
    bed = bed.with_columns(
        chromStart=(center - left).cast(pl.Int64),
        chromEnd=(center + right).cast(pl.Int64),
    )
    return bed


def _random_chain(
    *iterables: Iterable[T], seed: Optional[int] = None
) -> Generator[T, None, None]:
    """Chain iterables, randomly sampling from each until they are all exhausted."""
    rng = np.random.default_rng(seed)
    iterators = {i: iter(it) for i, it in enumerate(iterables)}
    while iterators:
        i = rng.choice(list(iterators.keys()))
        try:
            yield next(iterators[i])
        except StopIteration:
            del iterators[i]


def read_bedlike(path: Union[str, Path]) -> pl.DataFrame:
    """Reads a bed-like (i.e. `BED3+ <https://genome.ucsc.edu/FAQ/FAQformat.html#format1>`_) file as a
    polars DataFrame. The file type is inferred from the file extension. "Bed-like" refers to files
    with extension :code:`.bed`, :code:`.narrowPeak`, :code:`.broadPeak`, or otherwise a tabular (CSV, TSV, or feather/arrow)
    file with at least the BED3 columns. For tabular data, extra columns that are not part of the BED
    specification are allowed and kept in the resulting DataFrame.

    Parameters
    ----------
    path
        Path to the file.
    """
    path = Path(path)
    if ".bed" in path.suffixes:
        return _read_bed(path)
    elif ".narrowPeak" in path.suffixes:
        return _read_narrowpeak(path)
    elif ".broadPeak" in path.suffixes:
        return _read_broadpeak(path)
    else:
        try:
            return _read_bed_table(path)
        except ValueError:
            raise ValueError(
                f"""Unrecognized file extension: {"".join(path.suffixes)}. Expected one 
                of .bed, .narrowPeak, .broadPeak, or a table file (e.g. .csv, .tsv)"""
            )


class _BEDSchema(pa.DataFrameModel):
    chrom: pat.Series[str]
    chromStart: pat.Series[int]
    chromEnd: pat.Series[int]
    name: Optional[pat.Series[str]] = pa.Field(nullable=True)
    score: Optional[pat.Series[float]] = pa.Field(nullable=True)
    strand: Optional[pat.Series[str]] = pa.Field(isin=["+", "-", "."], nullable=True)
    thickStart: Optional[pat.Series[int]] = pa.Field(nullable=True)
    thickEnd: Optional[pat.Series[int]] = pa.Field(nullable=True)
    itemRgb: Optional[pat.Series[str]] = pa.Field(nullable=True)
    blockCount: Optional[pat.Series[pa.UInt]] = pa.Field(nullable=True)
    blockSizes: Optional[pat.Series[str]] = pa.Field(nullable=True)
    blockStarts: Optional[pat.Series[str]] = pa.Field(nullable=True)

    class Config:
        coerce = True


def _read_bed(bed_path: Union[str, Path]):
    with open(bed_path) as f:
        skip_rows = 0
        while (line := f.readline()).startswith(("track", "browser")):
            skip_rows += 1
    n_cols = line.count("\t") + 1
    bed_cols = [
        "chrom",
        "chromStart",
        "chromEnd",
        "name",
        "score",
        "strand",
        "thickStart",
        "thickEnd",
        "itemRgb",
        "blockCount",
        "blockSizes",
        "blockStarts",
    ]
    bed = pl.read_csv(
        bed_path,
        separator="\t",
        has_header=False,
        skip_rows=skip_rows,
        new_columns=bed_cols[:n_cols],
        schema_overrides={"chrom": pl.Utf8, "name": pl.Utf8, "strand": pl.Utf8},
        null_values=".",
    ).to_pandas()
    bed = _BEDSchema.to_schema()(bed)
    return pl.from_pandas(bed)


def _read_bed_table(table: Union[str, Path], **table_reader_kwargs):
    table = Path(table)
    suffixes = set(table.suffixes)
    reader_kwargs: Dict[str, Any] = {}
    reader_kwargs.update(table_reader_kwargs)
    if ".csv" in suffixes:
        reader_kwargs["separator"] = ","
        reader_kwargs["dtypes"] = {"chrom": pl.Utf8, "name": pl.Utf8, "strand": pl.Utf8}
        reader = pl.scan_csv
    elif {".txt", ".tsv"} & suffixes:
        reader_kwargs["separator"] = "\t"
        reader_kwargs["dtypes"] = {"chrom": pl.Utf8, "name": pl.Utf8, "strand": pl.Utf8}
        reader = pl.scan_csv
    elif {".fth", ".feather", ".ipc", ".arrow"} & suffixes:
        reader = pl.scan_ipc  # type : ignore[assignment]
    else:
        raise ValueError(f"Table has unrecognized file extension: {table.name}")
    bed = reader(table, **reader_kwargs).collect().to_pandas()
    bed = _BEDSchema.to_schema()(bed)
    return pl.from_pandas(bed)


class _NarrowPeakSchema(pa.DataFrameModel):
    chrom: pat.Series[str]
    chromStart: pat.Series[int]
    chromEnd: pat.Series[int]
    name: pat.Series[str] = pa.Field(nullable=True)
    score: pat.Series[float] = pa.Field(nullable=True)
    strand: pat.Series[str] = pa.Field(isin=["+", "-", "."], nullable=True)
    signalValue: pat.Series[float] = pa.Field(nullable=True)
    pValue: pat.Series[float] = pa.Field(nullable=True)
    qValue: pat.Series[float] = pa.Field(nullable=True)
    peak: pat.Series[int] = pa.Field(nullable=True)

    class Config:
        coerce = True


def _read_narrowpeak(narrowpeak_path: Union[str, Path]):
    with open(narrowpeak_path) as f:
        skip_rows = 0
        while f.readline().startswith(("track", "browser")):
            skip_rows += 1
    narrowpeaks = pl.read_csv(
        narrowpeak_path,
        separator="\t",
        has_header=False,
        skip_rows=skip_rows,
        new_columns=[
            "chrom",
            "chromStart",
            "chromEnd",
            "name",
            "score",
            "strand",
            "signalValue",
            "pValue",
            "qValue",
            "peak",
        ],
        schema_overrides={"chrom": pl.Utf8, "name": pl.Utf8, "strand": pl.Utf8},
        null_values=".",
    ).to_pandas()
    narrowpeaks = _NarrowPeakSchema.to_schema()(narrowpeaks)
    return pl.from_pandas(narrowpeaks)


class _BroadPeakSchema(pa.DataFrameModel):
    chrom: pat.Series[str]
    chromStart: pat.Series[int]
    chromEnd: pat.Series[int]
    name: pat.Series[str] = pa.Field(nullable=True)
    score: pat.Series[float] = pa.Field(nullable=True)
    strand: pat.Series[str] = pa.Field(isin=["+", "-", "."], nullable=True)
    signalValue: pat.Series[float] = pa.Field(nullable=True)
    pValue: pat.Series[float] = pa.Field(nullable=True)
    qValue: pat.Series[float] = pa.Field(nullable=True)

    class Config:
        coerce = True


def _read_broadpeak(broadpeak_path: Union[str, Path]):
    with open(broadpeak_path) as f:
        skip_rows = 0
        while f.readline().startswith(("track", "browser")):
            skip_rows += 1
    broadpeaks = pl.read_csv(
        broadpeak_path,
        separator="\t",
        has_header=False,
        skip_rows=skip_rows,
        new_columns=[
            "chrom",
            "chromStart",
            "chromEnd",
            "name",
            "score",
            "strand",
            "signalValue",
            "pValue",
            "qValue",
        ],
        schema_overrides={"chrom": pl.Utf8, "name": pl.Utf8, "strand": pl.Utf8},
        null_values=".",
    ).to_pandas()
    broadpeaks = _BroadPeakSchema.to_schema()(broadpeaks)
    return pl.from_pandas(broadpeaks)


def _cartesian_product(arrays: Sequence[NDArray]) -> NDArray:
    """Get the cartesian product of multiple arrays such that each entry corresponds to
    a unique combination of the input arrays' values.
    """
    # https://stackoverflow.com/a/49445693
    la = len(arrays)
    shape = *map(len, arrays), la
    dtype = np.result_type(*arrays)
    arr = np.empty(shape, dtype=dtype)
    arrs = (*accumulate(chain((arr,), repeat(0, la - 1)), np.ndarray.__getitem__),)
    idx = slice(None), *repeat(None, la - 1)
    for i in range(la - 1, 0, -1):
        arrs[i][..., i] = arrays[i][idx[: la - i]]
        arrs[i - 1][1:] = arrs[i]
    arr[..., 0] = arrays[0][idx]
    return arr.reshape(-1, la)


def _get_rel_starts(
    starts: NDArray[np.int64], ends: NDArray[np.int64]
) -> NDArray[np.int64]:
    rel_starts: NDArray[np.int64] = np.concatenate(
        [[0], (ends - starts).cumsum()[:-1]], dtype=np.int64
    )
    return rel_starts


DTYPE = TypeVar("DTYPE", bound=np.generic)


def _normalize_contig_name(contig: str, contigs: Iterable[str]) -> Optional[str]:
    """Normalize the contig name to adhere to the convention of the underlying file.
    i.e. remove or add "chr" to the contig name.

    Parameters
    ----------
    contig : str

    Returns
    -------
    str
        Normalized contig name.
    """
    for c in contigs:
        # exact match, remove chr, add chr
        if contig == c or contig[3:] == c or f"chr{contig}" == c:
            return c
    return None


ITYPE = TypeVar("ITYPE", bound=np.integer)


def _offsets_to_lengths(offsets: NDArray[ITYPE]) -> NDArray[ITYPE]:
    """Converts offsets to the number of elements in each group.

    Notes
    -----
    This function will silently fail with wraparound values if the offsets are
    not sorted in ascending order."""
    return np.diff(offsets)


def _lengths_to_offsets(
    lengths: NDArray[np.integer], dtype: type[ITYPE] = np.int64
) -> NDArray[ITYPE]:
    """Converts the number of elements in each group to a 1D array of offsets."""
    offsets = np.empty(lengths.size + 1, dtype=dtype)
    offsets[0] = 0
    np.cumsum(lengths, out=offsets[1:])
    return offsets


def idx_like_to_array(idx: Idx, max_len: int) -> NDArray[np.intp]:
    """Convert an index-like object to an array of non-negative indices. Shapes of multi-dimensional
    indices are preserved."""
    if isinstance(idx, slice):
        _idx = np.arange(max_len, dtype=np.intp)[idx]
    elif isinstance(idx, np.ndarray) and np.issubdtype(idx.dtype, np.bool_):
        _idx = idx.nonzero()[0]
    elif isinstance(idx, Sequence):
        _idx = np.asarray(idx, np.intp)
    else:
        _idx = idx

    if isinstance(_idx, (int, np.integer)):
        _idx = np.array([_idx], np.intp)

    # unable to type narrow from NDArray[bool] since it's a generic type
    _idx = cast(NDArray[np.intp], _idx)

    # handle negative indices
    _idx[_idx < 0] += max_len

    return _idx



from pathlib import Path
from typing import Sequence, Tuple, Union

import numpy as np
from numpy.typing import NDArray

__all__ = []

def count_intervals(
    paths: Sequence[Union[str, Path]],
    contig: str,
    starts: NDArray[np.int32],
    ends: NDArray[np.int32],
) -> NDArray[np.int32]:
    """Count how many intervals from BigWig files overlap with each region.

    Parameters
    ----------
    paths : Sequence[str | Path]
        Paths to BigWig files.
    contig : str
        Contig name.
    starts : NDArray[int32]
        Start positions.
    ends : NDArray[int32]
        End positions.

    Returns
    -------
    n_per_query : NDArray[int32]
        Shape = (regions, samples) Number of intervals per query.
    """
    ...

def intervals(
    paths: Sequence[Union[str, Path]],
    contig: str,
    starts: NDArray[np.int32],
    ends: NDArray[np.int32],
    offsets: NDArray[np.int64],
) -> Tuple[NDArray[np.uint32], NDArray[np.float32]]:
    """Load intervals from BigWig files.

    Parameters
    ----------
    paths : Sequence[str | Path]
        Paths to BigWig files.
    contig : str
        Contig name.
    starts : NDArray[int32]
        Start positions.
    ends : NDArray[int32]
        End positions.
    offsets : NDArray[int64]
        Offsets corresponding to the returned interval data of shape (regions, samples). Can be
        computed from the number of intervals per query, e.g. with the count_intervals function.

    Returns
    -------
    coordinates : NDArray[uint32]
        Shape = (intervals, 2) Coordinates.
    values : NDArray[float32]
        Shape = (intervals) Values.
    """
    ...



from pathlib import Path

import genvarloader as gvl
import numpy as np
import pytest
from pytest import fixture


@fixture
def pgen():
    pgen_path = Path(__file__).parent / "data" / "pgen" / "sample.pgen"
    pgen = gvl.Variants.from_file(pgen_path)
    return pgen


@pytest.mark.skip
def test_snp(pgen: gvl.Variants):
    contig = "19"
    start = np.array([110])  # VCF is 1-based
    end = np.array([111])

    true_geno = np.array(
        [
            [[0], [0]],
            [[0], [0]],
            [[0], [1]],
        ],
        dtype=np.int8,
    )

    gvl_geno = pgen.read(contig, start, end)

    assert gvl_geno is not None
    np.testing.assert_equal(gvl_geno.genotypes, true_geno)


@pytest.mark.skip
def test_del(pgen: gvl.Variants):
    contig = "19"
    start = np.array([1010695])
    end = np.array([1010696])

    true_positions = np.array([1010694, 1010695, 1010695], np.int32)
    true_size_diffs = np.array([-6, -3, -10], np.int32)
    true_ref = np.frombuffer(b"CGAGACGGAGAGAGACGGGGCC", "S1")
    true_ref_offsets = np.array([0, 7, 11, 22], np.uint32)
    true_alt = np.frombuffer(b"CGG", "S1")
    true_alt_offsets = np.array([0, 1, 2, 3], np.uint32)
    true_geno = np.array(
        [
            [[0, 1, 0], [0, 0, 1]],
            [[0, 0, 1], [0, 0, 1]],
            [[0, 0, 0], [1, 0, 0]],
        ],
        dtype=np.int8,
    )
    true_offsets = np.array([0, 3], np.uint32)

    gvl_geno = pgen.read(contig, start, end)
    gvl_hap_geno, max_ends = pgen.read_for_haplotype_construction(contig, start, end)

    assert gvl_geno is not None
    np.testing.assert_equal(gvl_geno.positions, true_positions)
    np.testing.assert_equal(gvl_geno.size_diffs, true_size_diffs)
    np.testing.assert_equal(gvl_geno.ref.alleles, true_ref)
    np.testing.assert_equal(gvl_geno.ref.offsets, true_ref_offsets)
    np.testing.assert_equal(gvl_geno.alt.alleles, true_alt)
    np.testing.assert_equal(gvl_geno.alt.offsets, true_alt_offsets)
    np.testing.assert_equal(gvl_geno.genotypes, true_geno)
    np.testing.assert_equal(gvl_geno.offsets, true_offsets)

    assert gvl_hap_geno is not None
    np.testing.assert_equal(gvl_hap_geno.positions, true_positions)
    np.testing.assert_equal(gvl_hap_geno.size_diffs, true_size_diffs)
    np.testing.assert_equal(gvl_hap_geno.ref.alleles, true_ref)
    np.testing.assert_equal(gvl_hap_geno.ref.offsets, true_ref_offsets)
    np.testing.assert_equal(gvl_hap_geno.alt.alleles, true_alt)
    np.testing.assert_equal(gvl_hap_geno.alt.offsets, true_alt_offsets)
    np.testing.assert_equal(gvl_hap_geno.genotypes, true_geno)
    np.testing.assert_equal(gvl_hap_geno.offsets, true_offsets)
    np.testing.assert_equal(max_ends, np.array([1010696 + 10], np.int64))


@pytest.mark.skip
def test_ins(pgen: gvl.Variants):
    contig = "19"
    start = np.array([1110695])
    end = np.array([1110696])

    true_positions = np.array([1110695, 1110695], np.int32)
    true_size_diffs = np.array([75, 0], np.int32)
    true_ref = np.frombuffer(b"AA", "S1")
    true_ref_offsets = np.array([0, 1, 2], np.uint32)
    true_alt = np.frombuffer(
        b"AGATAGATAGATAGATAGATAGATAGATAGATAGATAGATAGATAGATAGATAGATAGATAGATAGATAGATAGATG",
        "S1",
    )
    true_alt_offsets = np.array([0, 76, 77], np.uint32)
    true_geno = np.array(
        [
            [[0, 0], [1, 0]],
            [[1, 0], [1, 0]],
            [[0, 0], [0, 1]],
        ],
        dtype=np.int8,
    )
    true_offsets = np.array([0, 2], np.uint32)

    gvl_geno = pgen.read(contig, start, end)
    gvl_hap_geno, max_ends = pgen.read_for_haplotype_construction(contig, start, end)

    assert gvl_geno is not None
    np.testing.assert_equal(gvl_geno.positions, true_positions)
    np.testing.assert_equal(gvl_geno.size_diffs, true_size_diffs)
    np.testing.assert_equal(gvl_geno.ref.alleles, true_ref)
    np.testing.assert_equal(gvl_geno.ref.offsets, true_ref_offsets)
    np.testing.assert_equal(gvl_geno.alt.alleles, true_alt)
    np.testing.assert_equal(gvl_geno.alt.offsets, true_alt_offsets)
    np.testing.assert_equal(gvl_geno.genotypes, true_geno)
    np.testing.assert_equal(gvl_geno.offsets, true_offsets)

    assert gvl_hap_geno is not None
    np.testing.assert_equal(gvl_hap_geno.positions, true_positions)
    np.testing.assert_equal(gvl_hap_geno.size_diffs, true_size_diffs)
    np.testing.assert_equal(gvl_hap_geno.ref.alleles, true_ref)
    np.testing.assert_equal(gvl_hap_geno.ref.offsets, true_ref_offsets)
    np.testing.assert_equal(gvl_hap_geno.alt.alleles, true_alt)
    np.testing.assert_equal(gvl_hap_geno.alt.offsets, true_alt_offsets)
    np.testing.assert_equal(gvl_hap_geno.genotypes, true_geno)
    np.testing.assert_equal(gvl_hap_geno.offsets, true_offsets)
    np.testing.assert_equal(max_ends, end)


@pytest.mark.skip
def test_split_snp(pgen: gvl.Variants):
    contig = "20"
    start = np.array([1110695])  # VCF is 1-based
    end = np.array([1110696])

    true_positions = np.array([1110695, 1110695], np.int32)
    true_size_diffs = np.array([0, 0], np.int32)
    true_ref = np.array([b"G", b"G"])
    true_alt = np.array([b"A", b"T"])
    true_geno = np.array(
        [
            [[1, 0], [0, 1]],
            [[0, 1], [1, 0]],
            [[-9, 1], [-9, 1]],
        ],
        dtype=np.int8,
    )
    true_offsets = np.array([0, 2], np.uint32)

    gvl_geno = pgen.read(contig, start, end)
    gvl_hap_geno, max_ends = pgen.read_for_haplotype_construction(contig, start, end)

    assert gvl_geno is not None
    np.testing.assert_equal(gvl_geno.positions, true_positions)
    np.testing.assert_equal(gvl_geno.size_diffs, true_size_diffs)
    np.testing.assert_equal(gvl_geno.ref.alleles, true_ref)
    np.testing.assert_equal(gvl_geno.alt.alleles, true_alt)
    np.testing.assert_equal(gvl_geno.genotypes, true_geno)
    np.testing.assert_equal(gvl_geno.offsets, true_offsets)


if __name__ == "__main__":
    # test_del(pgen())
    test_split_snp(pgen())



import numpy as np
from attrs import define
from einops import repeat
from genvarloader._dataset._intervals import intervals_to_tracks, tracks_to_intervals
from genvarloader._ragged import INTERVAL_DTYPE, RaggedIntervals
from numpy.typing import NDArray
from pytest_cases import parametrize_with_cases


@define
class Data:
    interval_idx: NDArray[np.intp]
    regions: NDArray[np.int32]
    intervals: RaggedIntervals
    track: NDArray[np.float32]
    t_offsets: NDArray[np.int64]


def case_simple():
    """Return intervals and values. After decompressing, should yield:
    [1, 1, 1, 0, 2, 0, 0, 0, 0, 3]
    """
    coordinates = np.array(
        [
            [0, 3],
            [3, 4],
            [4, 5],
            [5, 9],
            [9, 10],
        ],
        dtype=np.int32,
    )
    values = np.array([1, 0, 2, 0, 3], dtype=np.float32)
    n_intervals = len(values)
    intervals = np.empty(n_intervals, dtype=INTERVAL_DTYPE)
    intervals["start"] = coordinates[:, 0]
    intervals["end"] = coordinates[:, 1]
    intervals["value"] = values
    intervals = RaggedIntervals.from_lengths(
        intervals, np.array([n_intervals], np.int32)
    )

    interval_idx = np.array([0], dtype=np.intp)

    regions = np.array([[0, 0, 10]], dtype=np.int32)  # coordinates 0:0-10

    track = np.array([1, 1, 1, 0, 2, 0, 0, 0, 0, 3], dtype=np.float32)
    t_offsets = np.array([0, len(track)], np.int64)
    return Data(interval_idx, regions, intervals, track, t_offsets)


def case_two_regions():
    """Return intervals and values. After decompressing, should yield:
    [1, 1, 1, 0, 2, 0, 0, 0, 0, 3]
    """
    coordinates = np.array(
        [
            [0, 3],
            [3, 4],
            [4, 5],
            [5, 9],
            [9, 10],
        ],
        dtype=np.int32,
    )
    values = np.array([1, 0, 2, 0, 3], dtype=np.float32)
    n_intervals = len(values)

    intervals = np.empty(n_intervals, dtype=INTERVAL_DTYPE)
    intervals["start"] = coordinates[:, 0]
    intervals["end"] = coordinates[:, 1]
    intervals["value"] = values
    intervals = repeat(intervals, "n -> (r n)", r=2)

    intervals = RaggedIntervals.from_lengths(
        intervals, np.array([n_intervals, n_intervals], np.int32)
    )

    interval_idx = np.array([0, 1], dtype=np.intp)

    regions = np.array([[0, 0, 10]], dtype=np.int32)  # coordinates 0:0-10
    regions = repeat(regions, "n d -> (r n) d", r=2)

    track = np.array([1, 1, 1, 0, 2, 0, 0, 0, 0, 3], dtype=np.float32)
    tracks = repeat(track, "n -> (r n)", r=2)
    t_offsets = np.array([0, len(track), tracks.size], np.int64)
    return Data(interval_idx, regions, intervals, tracks, t_offsets)


@parametrize_with_cases("data", cases=".")
def test_intervals_to_tracks(data: Data):
    intervals = data.intervals
    out = np.empty_like(data.track)
    out_offsets = data.t_offsets.copy()
    intervals_to_tracks(
        offset_idxs=data.interval_idx,
        starts=data.regions[:, 1],
        intervals=intervals.data,
        itv_offsets=intervals.offsets,
        out=out,
        out_offsets=out_offsets,
    )
    np.testing.assert_array_equal(out, data.track)


@parametrize_with_cases("data", cases=".")
def test_tracks_to_intervals(data: Data):
    intervals, offsets = tracks_to_intervals(data.regions, data.track, data.t_offsets)
    intervals = RaggedIntervals.from_offsets(intervals, 1, offsets)
    np.testing.assert_array_equal(intervals.data, data.intervals.data)
    np.testing.assert_array_equal(intervals.offsets, data.intervals.offsets)



import subprocess
from pathlib import Path
from textwrap import dedent
from typing import Annotated, List

import typer

WDIR = Path(__file__).resolve().parent
SEQ_LEN = 20


def run_shell(cmd: List[str], input=None):
    try:
        prc = subprocess.run(cmd, check=True, capture_output=True, input=input)
    except subprocess.CalledProcessError as e:
        print("Command:", " ".join(e.cmd))
        print("Stdout:", e.stdout.decode())
        print("Error message:", e.stderr.decode())
        raise e
    return prc


def main(
    name: Annotated[
        str,
        typer.Argument(
            help=dedent(
                """
            Prefix for files. There should be a file named <name>.vcf and then this
            script will generate <name>.bed and <name>_<sample>_nr<row_nr>_h<hap_nr>.fa
            files.
            """
            ),
        ),
    ] = "sample",
    reference: Annotated[
        Path,
        typer.Argument(
            help="Path to reference genome.",
        ),
    ] = WDIR / "fasta" / "Homo_sapiens.GRCh38.dna.primary_assembly.fa.bgz",
    out_dir: Path = WDIR / "consensus",
    indels: Annotated[bool, typer.Option(help="Whether to include indels.")] = True,
    structural_variants: Annotated[
        bool, typer.Option(help="Whether to include structural variants.")
    ] = False,
    multiallelic: Annotated[
        bool, typer.Option(help="Whether to allow multiallelic variants.")
    ] = False,
):
    """Generate ground truth variant sequences using `bcftools consensus`."""
    import shutil
    import sys
    from time import perf_counter

    import genvarloader as gvl
    import polars as pl
    import polars.selectors as cs
    from loguru import logger
    from tqdm.auto import tqdm

    log_file = Path(__file__).parent / "generate_ground_truth.log"
    log_file.unlink()
    logger.enable("genvarloader")
    logger.remove(0)
    logger.add(sys.stdout, level="INFO")
    logger.add(log_file, level="DEBUG")

    logger.info(
        f"""Running command:
        generate_ground_truth.py {name} {reference} {out_dir} --indels {indels} --structural-variants {structural_variants} --multiallelic {multiallelic}
        """
    )

    t0 = perf_counter()

    out_dir = out_dir.resolve()
    if out_dir.exists():
        shutil.rmtree(out_dir)
    out_dir.mkdir(0o777, parents=True, exist_ok=True)
    vcf_path = WDIR / "vcf" / f"{name}.vcf"
    filtered_vcf = WDIR / "vcf" / f"filtered_{name}.vcf"

    with open(vcf_path, "r") as f:
        vcf = f.read().encode()
    # left-align
    norm_cmd = [
        "bcftools",
        "norm",
        "-f",
        str(reference),
    ]
    result = run_shell(norm_cmd, input=vcf)
    logger.info(f"Left-alignment: {result.stderr.decode()}")
    vcf = result.stdout

    if not multiallelic:
        logger.info("Splitting multiallelic variants.")
        split_multiallelics_cmd = [
            "bcftools",
            "norm",
            "-a",
            "--atom-overlaps",
            ".",
            "-f",
            str(reference),
            "-m",
            "-",
        ]
        result = run_shell(split_multiallelics_cmd, input=vcf)
        logger.info(
            f"Atomizing variants and splitting multiallelics: {result.stderr.decode()}"
        )
        vcf = result.stdout
    if not indels:
        logger.info("Ignoring indels.")
        remove_indel_cmd = [
            "bcftools",
            "view",
            "-e",
            'TYPE="indel"',
        ]
        vcf = run_shell(remove_indel_cmd, input=vcf).stdout
    if not structural_variants:
        logger.info("Ignoring structural variants.")
        remove_sv_cmd = [
            "bcftools",
            "view",
            "-e",
            'TYPE="OTHER"',
        ]
        vcf = run_shell(remove_sv_cmd, input=vcf).stdout
    with open(filtered_vcf, "w+t") as f:
        f.write(vcf.decode())

    # BGZIP the VCF or bcftools consensus errors out
    run_shell(
        [
            "bcftools",
            "view",
            "-O",
            "z",
            "-o",
            str(filtered_vcf.with_suffix(".vcf.gz")),
            str(filtered_vcf),
        ]
    )
    filtered_vcf = filtered_vcf.with_suffix(".vcf.gz")
    run_shell(["bcftools", "index", str(filtered_vcf)])

    logger.info("Generating PGEN file.")
    run_shell(
        [
            "plink2",
            "--vcf",
            str(filtered_vcf),
            "--make-pgen",
            "--vcf-half-call",
            "r",
            "--out",
            str(WDIR / "pgen" / f"filtered_{name}"),
        ]
    )

    bed = pl.read_csv(
        filtered_vcf,
        separator="\t",
        comment_prefix="#",
        has_header=False,
        new_columns=[
            "chrom",
            "pos",
            "ID",
            "REF",
            "ALT",
            "QUAL",
            "FILTER",
            "INFO",
            "FORMAT",
            "NA00001",
            "NA00002",
            "NA00003",
        ],
    )
    samples = bed.select(cs.matches(r"^NA\d{5}$")).columns

    #! when writing the source VCF, make sure that the most distant variants within a group are not more than SEQ_LEN // 2 apart
    #! also ensure that groups of variants are more than SEQ_LEN apart
    #! otherwise, the logic below won't work
    bed = (
        bed.group_by("chrom", maintain_order=True)
        .agg(
            "pos",
            (pl.col("pos").diff().fill_null(0) > SEQ_LEN).cum_sum().alias("group"),
        )
        .explode("pos", "group")
        .group_by("chrom", "group", maintain_order=True)
        .agg(
            start=pl.col("pos").min() - SEQ_LEN // 2,
            end=pl.col("pos").min() + SEQ_LEN // 2,
        )
        .drop("group")
        .sample(fraction=1, shuffle=True, seed=0)
        .with_row_index()
    )

    logger.info("Generating BED file.")
    (
        bed.select("chrom", "start", "end").write_csv(
            WDIR / "vcf" / f"{name}.bed", include_header=False, separator="\t"
        )
    )

    logger.info("Generating consensus sequences.")
    pbar = tqdm(total=bed.height * len(samples) * 2)
    for row in bed.select("index", "chrom", "start", "end").iter_rows():
        row_nr, chrom, start, end = row
        subseq_cmd = [
            "samtools",
            "faidx",
            str(reference),
            f"{chrom}:{start + 1}-{end}",
        ]
        for sample in samples:
            for hap in range(2):
                out_fasta = out_dir / f"{name}_{sample}_nr{row_nr}_h{hap}.fa"
                consensus_cmd = [
                    "bcftools",
                    "consensus",
                    "-H",
                    str(hap + 1),
                    "-s",
                    sample,
                    "-o",
                    str(out_fasta),
                    str(filtered_vcf),
                ]
                seq = run_shell(subseq_cmd)
                run_shell(consensus_cmd, input=seq.stdout)
                index_cmd = ["samtools", "faidx", str(out_fasta)]
                run_shell(index_cmd)
                pbar.update()
    pbar.close()

    bed = WDIR / "vcf" / f"{name}.bed"

    logger.info("Generating phased dataset.")
    gvl.write(
        path=WDIR / "phased_dataset.gvl",
        bed=bed,
        variants=filtered_vcf,
        overwrite=True,
    )

    logger.info("Generating unphased dataset.")
    gvl.write(
        path=WDIR / "unphased_dataset.gvl",
        bed=bed,
        variants=filtered_vcf,
        overwrite=True,
        phased=False,
        dosage_field="VAF",
    )

    logger.info(f"Finished in {perf_counter() - t0} seconds.")


if __name__ == "__main__":
    typer.run(main)



from pathlib import Path

import numpy as np
import pytest
from genvarloader._fasta import Fasta, NoPadError
from pysam import FastaFile
from pytest import fixture


@fixture
def fasta_path():
    return (
        Path(__file__).parent
        / "data"
        / "fasta"
        / "Homo_sapiens.GRCh38.dna.primary_assembly.fa.bgz"
    )


def test_pad_right(fasta_path):
    fasta = Fasta("ref", fasta_path, pad="N")
    contig = "1"
    with FastaFile(fasta_path) as f:
        end_of_contig = f.get_reference_length(contig)
        start = end_of_contig - 5
        end = start + 10
        desired = np.full(end - start, b"N", "S1")
        desired[:5] = np.frombuffer(
            f.fetch(contig, start, start + 5).encode("ascii"), "S1"
        )

    seq = fasta.read(contig, start, end)

    np.testing.assert_equal(seq, desired)


def test_pad_left(fasta_path):
    fasta = Fasta("ref", fasta_path, pad="N")
    contig = "1"
    start = -5
    end = start + 10
    seq = fasta.read(contig, start, end)
    with FastaFile(fasta_path) as f:
        desired = np.full(end - start, b"N", "S1")
        desired[5:] = np.frombuffer(f.fetch(contig, 0, end).encode("ascii"), "S1")

    np.testing.assert_equal(seq, desired)


def test_no_pad(fasta_path):
    fasta = Fasta("ref", fasta_path)
    end_of_contig_1 = 248956422
    contig = "1"
    start = end_of_contig_1 - 5
    end = start + 10
    with pytest.raises(NoPadError):
        fasta.read(contig, start, end)

    contig = "1"
    start = -5
    end = start + 10
    with pytest.raises(NoPadError):
        fasta.read(contig, start, end)



from __future__ import annotations

from pathlib import Path
from typing import (
    TYPE_CHECKING,
    Callable,
    Dict,
    List,
    Literal,
    Optional,
    Protocol,
    Tuple,
    TypeVar,
    Union,
    cast,
    overload,
)

import numba as nb
import numpy as np
import polars as pl
from attrs import define, evolve
from einops import repeat
from loguru import logger
from numpy.typing import NDArray
from tqdm.auto import tqdm

from .._fasta import Fasta
from .._ragged import (
    INTERVAL_DTYPE,
    Ragged,
    RaggedAnnotatedHaps,
    RaggedIntervals,
)
from .._utils import _lengths_to_offsets, _normalize_contig_name
from .._variants._records import VLenAlleles
from ._genotypes import (
    SparseGenotypes,
    SparseSomaticGenotypes,
    choose_unphased_variants,
    get_diffs_sparse,
    reconstruct_haplotypes_from_sparse,
)
from ._indexing import DatasetIndexer
from ._intervals import intervals_to_tracks, tracks_to_intervals
from ._tracks import shift_and_realign_tracks_sparse
from ._utils import padded_slice, splits_sum_le_value

T = TypeVar("T", covariant=True)


@define
class Reference:
    reference: NDArray[np.uint8]
    contigs: List[str]
    offsets: NDArray[np.uint64]
    pad_char: int

    @classmethod
    def from_path_and_contigs(cls, fasta: Union[str, Path], contigs: List[str]):
        _fasta = Fasta("ref", fasta, "N")

        if not _fasta.cache_path.exists():
            logger.info("Memory-mapping FASTA file for faster access.")
            _fasta._write_to_cache()

        contigs = cast(
            List[str],
            [_normalize_contig_name(c, _fasta.contigs) for c in contigs],
        )
        _fasta.sequences = _fasta._get_sequences(contigs)
        if TYPE_CHECKING:
            assert _fasta.sequences is not None
            assert _fasta.pad is not None
        refs: List[NDArray[np.bytes_]] = []
        next_offset = 0
        _ref_offsets: Dict[str, int] = {}
        for contig in contigs:
            arr = _fasta.sequences[contig]
            refs.append(arr)
            _ref_offsets[contig] = next_offset
            next_offset += len(arr)
        reference = np.concatenate(refs).view(np.uint8)
        pad_char = ord(_fasta.pad)
        if any(c is None for c in contigs):
            raise ValueError("Contig names in metadata do not match reference.")
        ref_offsets = np.empty(len(contigs) + 1, np.uint64)
        ref_offsets[:-1] = np.array([_ref_offsets[c] for c in contigs], dtype=np.uint64)
        ref_offsets[-1] = len(reference)
        return cls(reference, contigs, ref_offsets, pad_char)


@define
class _Variants:
    positions: NDArray[np.int32]
    sizes: NDArray[np.int32]
    alts: VLenAlleles

    @classmethod
    def from_table(cls, variants: Union[str, Path, pl.DataFrame]):
        if isinstance(variants, (str, Path)):
            variants = pl.read_ipc(variants)
        return cls(
            variants["POS"].to_numpy(),
            variants["ILEN"].to_numpy(),
            VLenAlleles.from_polars(variants["ALT"]),
        )


class Reconstructor(Protocol[T]):
    """Reconstructs data on-the-fly. e.g. personalized sequences, tracks, etc."""

    def __call__(
        self,
        idx: NDArray[np.integer],
        regions: NDArray[np.int32],
        output_length: Union[Literal["ragged", "variable"], int],
        jitter: int,
        rng: Optional[np.random.Generator],
    ) -> T: ...


@define
class Seqs(Reconstructor[Ragged[np.bytes_]]):
    reference: Reference
    """The reference genome. This is kept in memory."""

    def __call__(
        self,
        idx: NDArray[np.integer],
        regions: NDArray[np.int32],
        output_length: Union[Literal["ragged", "variable"], int],
        jitter: int,
        rng: Optional[np.random.Generator],
    ) -> Ragged[np.bytes_]:
        batch_size = len(idx)
        lengths = regions[:, 2] - regions[:, 1]

        if not isinstance(output_length, int):
            # (b)
            out_lengths = lengths
        else:
            out_lengths = np.full(
                batch_size, output_length + 2 * jitter, dtype=np.int32
            )

        if rng is not None and isinstance(output_length, int):
            # (b)
            max_shift = (lengths - output_length).clip(min=0)
            shifts = rng.integers(0, max_shift + 1, dtype=np.int32)
            regions = regions.copy()
            regions[:, 1] += shifts
            regions[:, 2] = regions[:, 1] + output_length + 2 * jitter

        # (b+1)
        out_offsets = _lengths_to_offsets(out_lengths)

        # ragged (b)
        ref = _get_reference(
            regions=regions,
            out_offsets=out_offsets,
            reference=self.reference.reference,
            ref_offsets=self.reference.offsets,
            pad_char=self.reference.pad_char,
        ).view("S1")
        ref = cast(Ragged[np.bytes_], Ragged.from_offsets(ref, batch_size, out_offsets))

        return ref


@nb.njit(parallel=True, nogil=True, cache=True)
def _get_reference(
    regions: NDArray[np.int32],
    out_offsets: NDArray[np.int64],
    reference: NDArray[np.uint8],
    ref_offsets: NDArray[np.uint64],
    pad_char: int,
) -> NDArray[np.uint8]:
    out = np.empty(out_offsets[-1], np.uint8)
    for i in nb.prange(len(regions)):
        o_s, o_e = out_offsets[i], out_offsets[i + 1]
        c_idx, start, end = regions[i, :3]
        c_s = ref_offsets[c_idx]
        c_e = ref_offsets[c_idx + 1]
        out[o_s:o_e] = padded_slice(reference[c_s:c_e], start, end, pad_char)
    return out


H = TypeVar("H", Ragged[np.bytes_], RaggedAnnotatedHaps)


@define
class Haps(Reconstructor[H]):
    reference: Reference
    """The reference genome. This is kept in memory."""
    variants: _Variants
    """The variant sites in the dataset. This is kept in memory."""
    genotypes: Union[SparseGenotypes, SparseSomaticGenotypes]
    """Shape: (regions, samples, ploidy). The genotypes in the dataset. This is memory mapped."""
    annotate: bool

    @classmethod
    def from_path(
        cls: type[Haps[Ragged[np.bytes_]]],
        path: Path,
        reference: Reference,
        phased: bool,
        regions: NDArray[np.int32],
        samples: List[str],
        ploidy: int,
    ) -> Haps[Ragged[np.bytes_]]:
        variants = _Variants.from_table(path / "genotypes" / "variants.arrow")
        if phased:
            genotypes = SparseGenotypes(
                np.memmap(
                    path / "genotypes" / "variant_idxs.npy",
                    dtype=np.int32,
                    mode="r",
                ),
                np.memmap(path / "genotypes" / "offsets.npy", dtype=np.int64, mode="r"),
                len(regions),
                len(samples),
                ploidy,
            )
        else:
            genotypes = SparseSomaticGenotypes(
                np.memmap(
                    path / "genotypes" / "variant_idxs.npy",
                    dtype=np.int32,
                    mode="r",
                ),
                np.memmap(
                    path / "genotypes" / "ccfs.npy", dtype=np.float32, mode="r"
                ),
                np.memmap(path / "genotypes" / "offsets.npy", dtype=np.int64, mode="r"),
                len(regions),
                len(samples),
            )
        return cls(
            reference=reference,
            variants=variants,
            genotypes=genotypes,
            annotate=False,
        )

    def _haplotype_ilens(
        self,
        idx: NDArray[np.integer],
        jittered_regions: NDArray[np.int32],
        deterministic: bool,
        keep: NDArray[np.bool_] | None = None,
        keep_offsets: NDArray[np.int64] | None = None,
    ) -> NDArray[np.int32]:
        """`idx` must be 1D."""
        # (b p)
        geno_offset_idxs = self.get_geno_offset_idx(idx, self.genotypes)

        if isinstance(self.genotypes, SparseSomaticGenotypes):
            if keep is None or keep_offsets is None:
                keep, keep_offsets = choose_unphased_variants(
                    starts=jittered_regions[:, 1],
                    ends=jittered_regions[:, 2],
                    geno_offset_idxs=geno_offset_idxs,
                    geno_v_idxs=self.genotypes.variant_idxs,
                    geno_offsets=self.genotypes.offsets,
                    positions=self.variants.positions,
                    sizes=self.variants.sizes,
                    ccfs=self.genotypes.ccfs,
                    deterministic=deterministic,
                )
            # (r s p)
            hap_ilens = get_diffs_sparse(
                geno_offset_idxs=geno_offset_idxs,
                geno_v_idxs=self.genotypes.variant_idxs,
                geno_offsets=self.genotypes.offsets,
                size_diffs=self.variants.sizes,
                keep=keep,
                keep_offsets=keep_offsets,
            )
        else:
            # (r s p)
            hap_ilens = get_diffs_sparse(
                geno_offset_idxs=geno_offset_idxs,
                geno_v_idxs=self.genotypes.variant_idxs,
                geno_offsets=self.genotypes.offsets,
                size_diffs=self.variants.sizes,
                starts=jittered_regions[:, 1],
                ends=jittered_regions[:, 2],
                positions=self.variants.positions,
            )

        return hap_ilens.reshape(-1, self.genotypes.ploidy)

    @overload
    def with_annot(self, annotations: Literal[False]) -> Haps[Ragged[np.bytes_]]: ...
    @overload
    def with_annot(self, annotations: Literal[True]) -> Haps[RaggedAnnotatedHaps]: ...

    def with_annot(self, annotations: bool) -> Haps:
        return evolve(self, annotate=annotations)

    def __call__(
        self,
        idx: NDArray[np.integer],
        regions: NDArray[np.int32],
        output_length: Union[Literal["ragged", "variable"], int],
        jitter: int,
        rng: Optional[np.random.Generator],
    ) -> H:
        haps, *_ = self.get_haps_and_shifts(
            idx=idx,
            regions=regions,
            output_length=output_length,
            jitter=jitter,
            rng=rng,
        )
        return haps

    def get_haps_and_shifts(
        self,
        idx: NDArray[np.integer],
        regions: NDArray[np.int32],
        output_length: Union[Literal["ragged", "variable"], int],
        jitter: int,
        rng: Optional[np.random.Generator],
    ) -> tuple[
        H,
        NDArray[np.intp],
        NDArray[np.int32],
        NDArray[np.int32],
        NDArray[np.int32],
        Optional[NDArray[np.bool_]],
        Optional[NDArray[np.int64]],
    ]:
        batch_size = len(idx)
        lengths = regions[:, 2] - regions[:, 1]

        geno_offset_idx = self.get_geno_offset_idx(idx, self.genotypes)

        if isinstance(self.genotypes, SparseSomaticGenotypes):
            keep, keep_offsets = choose_unphased_variants(
                starts=regions[:, 1],
                ends=regions[:, 2],
                geno_offset_idxs=geno_offset_idx,
                geno_v_idxs=self.genotypes.variant_idxs,
                geno_offsets=self.genotypes.offsets,
                ccfs=self.genotypes.ccfs,
                positions=self.variants.positions,
                sizes=self.variants.sizes,
                deterministic=rng is None,
            )
        else:
            keep = None
            keep_offsets = None

        # (b p)
        diffs = self._haplotype_ilens(idx, regions, rng is None, keep, keep_offsets)
        hap_lengths = lengths[:, None] + diffs

        if rng is None or isinstance(output_length, str):
            # (b p)
            shifts = np.zeros((batch_size, self.genotypes.ploidy), dtype=np.int32)
        else:
            # if the haplotype is longer than the region, shift it randomly
            # by up to:
            # the difference in length between the haplotype and the region
            # PLUS the difference in length between the region and the output_length
            # (b p)
            max_shift = diffs.clip(min=0)
            if isinstance(output_length, int):
                # (b p)
                max_shift += (lengths - output_length).clip(min=0)[:, None]
            shifts = rng.integers(0, max_shift + 1, dtype=np.int32)

        if not isinstance(output_length, int):
            # (b p)
            out_lengths = hap_lengths
        else:
            out_lengths = np.full(
                (batch_size, self.genotypes.ploidy),
                output_length + 2 * jitter,
                dtype=np.int32,
            )
        # (b*p+1)
        out_offsets = _lengths_to_offsets(out_lengths)

        # (b p l), (b p l), (b p l)
        if self.annotate:
            haps, maybe_annot_v_idx, maybe_annot_pos = self._get_haplotypes(
                geno_offset_idx=geno_offset_idx,
                regions=regions,
                out_offsets=out_offsets,
                shifts=shifts,
                keep=keep,
                keep_offsets=keep_offsets,
                annotate=self.annotate,
            )
        else:
            haps = self._get_haplotypes(
                geno_offset_idx=geno_offset_idx,
                regions=regions,
                out_offsets=out_offsets,
                shifts=shifts,
                keep=keep,
                keep_offsets=keep_offsets,
                annotate=self.annotate,
            )

        if isinstance(self.genotypes, SparseSomaticGenotypes):
            # (b 1 l) -> (b l) remove ploidy dim
            haps = haps.squeeze(1)

        if self.annotate:
            out = RaggedAnnotatedHaps(haps, maybe_annot_v_idx, maybe_annot_pos)  # type: ignore
        else:
            out = haps

        return (
            out,  # type: ignore | pylance doesn't like this but it's correct behavior for the signature
            geno_offset_idx,
            shifts,
            diffs,
            hap_lengths,
            keep,
            keep_offsets,
        )

    @staticmethod
    def get_geno_offset_idx(
        idx: NDArray[np.integer],
        genotypes: Union[SparseGenotypes, SparseSomaticGenotypes],
    ) -> NDArray[np.intp]:
        r_idx, s_idx = np.unravel_index(idx, genotypes.effective_shape[:2])
        ploid_idx = np.arange(genotypes.ploidy, dtype=np.intp)
        rsp_idx = (r_idx[:, None], s_idx[:, None], ploid_idx)
        geno_offset_idx = np.ravel_multi_index(rsp_idx, genotypes.effective_shape)
        return geno_offset_idx

    @overload
    def _get_haplotypes(
        self,
        geno_offset_idx: NDArray[np.intp],
        regions: NDArray[np.int32],
        out_offsets: NDArray[np.int64],
        shifts: NDArray[np.int32],
        keep: Optional[NDArray[np.bool_]],
        keep_offsets: Optional[NDArray[np.int64]],
        annotate: Literal[False],
    ) -> Ragged[np.bytes_]: ...
    @overload
    def _get_haplotypes(
        self,
        geno_offset_idx: NDArray[np.intp],
        regions: NDArray[np.int32],
        out_offsets: NDArray[np.int64],
        shifts: NDArray[np.int32],
        keep: Optional[NDArray[np.bool_]],
        keep_offsets: Optional[NDArray[np.int64]],
        annotate: Literal[True],
    ) -> Tuple[Ragged[np.bytes_], Ragged[np.int32], Ragged[np.int32]]: ...

    def _get_haplotypes(
        self,
        geno_offset_idx: NDArray[np.intp],
        regions: NDArray[np.int32],
        out_offsets: NDArray[np.int64],
        shifts: NDArray[np.int32],
        keep: Optional[NDArray[np.bool_]],
        keep_offsets: Optional[NDArray[np.int64]],
        annotate: bool,
    ) -> (
        Ragged[np.bytes_] | Tuple[Ragged[np.bytes_], Ragged[np.int32], Ragged[np.int32]]
    ):
        """Reconstruct haplotypes from sparse genotypes.

        Parameters
        ----------
        geno_offset_idx
            Shape: (queries). The genotype offset indices. i.e. the dataset indices.
        regions
            Shape: (queries). The regions to reconstruct.
        out_offsets
            Shape: (queries+1). Offsets for haplotypes and annotations.
        shifts
            Shape: (queries, ploidy). The shift for each haplotype.
        keep
            Ragged array, shape: (variants). Whether to keep each variant. Implicitly has the same offsets
            as the sparse genotypes corresponding to geno_offset_idx.
        """
        haps = Ragged.from_offsets(
            np.empty(out_offsets[-1], np.uint8), shifts.shape, out_offsets
        )

        if annotate:
            annot_v_idxs = Ragged.from_offsets(
                np.empty(out_offsets[-1], np.int32), shifts.shape, out_offsets
            )
            annot_positions = Ragged.from_offsets(
                np.empty(out_offsets[-1], np.int32), shifts.shape, out_offsets
            )
        else:
            annot_v_idxs = None
            annot_positions = None

        # don't need to pass annot offsets because they are the same as haps offsets
        reconstruct_haplotypes_from_sparse(
            geno_offset_idxs=geno_offset_idx,
            out=haps.data,
            out_offsets=haps.offsets,
            regions=regions,
            shifts=shifts,
            geno_offsets=self.genotypes.offsets,
            geno_v_idxs=self.genotypes.variant_idxs,
            positions=self.variants.positions,
            sizes=self.variants.sizes,
            alt_alleles=self.variants.alts.alleles.view(np.uint8),
            alt_offsets=self.variants.alts.offsets,
            ref=self.reference.reference,
            ref_offsets=self.reference.offsets,
            pad_char=self.reference.pad_char,
            keep=keep,
            keep_offsets=keep_offsets,
            annot_v_idxs=annot_v_idxs.data
            if annot_v_idxs is not None
            else annot_v_idxs,
            annot_ref_pos=annot_positions.data
            if annot_positions is not None
            else annot_positions,
        )

        haps.data = haps.data.view("S1")
        haps = cast(Ragged[np.bytes_], haps)

        if annotate:
            return haps, annot_v_idxs, annot_positions  # type: ignore
        else:
            return haps


@define
class Tracks(Reconstructor[Ragged[np.float32]]):
    intervals: Dict[str, RaggedIntervals]
    """The intervals in the dataset. This is memory mapped."""
    active_tracks: List[str]

    def with_tracks(self, tracks: str | List[str]) -> Tracks:
        if isinstance(tracks, str):
            tracks = [tracks]
        if missing := list(set(tracks) - set(self.intervals)):
            raise ValueError(f"Missing tracks: {missing}")
        return evolve(self, active_tracks=tracks)

    @classmethod
    def from_path(cls, path: Path, regions: NDArray[np.int32], n_samples: int):
        available_tracks: List[str] = []
        for p in (path / "intervals").iterdir():
            if len(list(p.iterdir())) == 0:
                p.rmdir()
            else:
                available_tracks.append(p.name)
        available_tracks.sort()
        intervals: Optional[Dict[str, RaggedIntervals]] = {}
        for track in available_tracks:
            itvs = np.memmap(
                path / "intervals" / track / "intervals.npy",
                dtype=INTERVAL_DTYPE,
                mode="r",
            )
            offsets = np.memmap(
                path / "intervals" / track / "offsets.npy",
                dtype=np.int64,
                mode="r",
            )
            intervals[track] = RaggedIntervals.from_offsets(
                itvs, (len(regions), n_samples), offsets
            )
        return cls(intervals, available_tracks)

    def __call__(
        self,
        idx: NDArray[np.integer],
        regions: NDArray[np.int32],
        output_length: Union[Literal["ragged", "variable"], int],
        jitter: int,
        rng: Optional[np.random.Generator],
    ) -> Ragged[np.float32]:
        lengths = regions[:, 2] - regions[:, 1]
        if isinstance(output_length, int):
            out_lengths = track_lengths = np.full_like(
                lengths, output_length + 2 * jitter
            )
            if rng is not None:
                max_shift = (lengths - output_length).clip(min=0)
                shifts = rng.integers(0, max_shift + 1, dtype=np.int32)
                regions = regions.copy()
                regions[:, 1] += shifts
                regions[:, 2] = regions[:, 1] + output_length + 2 * jitter
        else:
            out_lengths = track_lengths = lengths

        # (b [p])
        out_ofsts_per_t = _lengths_to_offsets(out_lengths)
        track_ofsts_per_t = _lengths_to_offsets(track_lengths)
        # caller accounts for ploidy
        n_per_track = out_ofsts_per_t[-1]
        # ragged (b t [p] l)
        out = np.empty(len(self.active_tracks) * n_per_track, np.float32)
        out_lens = repeat(out_lengths, "b -> b t", t=len(self.active_tracks))
        out_offsets = _lengths_to_offsets(out_lens)

        for track_ofst, name in enumerate(self.active_tracks):
            intervals = self.intervals[name]
            # (b t l) ragged
            _out = out[track_ofst * n_per_track : (track_ofst + 1) * n_per_track]
            intervals_to_tracks(
                offset_idxs=idx,
                starts=regions[:, 1],
                intervals=intervals.data,
                itv_offsets=intervals.offsets,
                out=_out,
                out_offsets=track_ofsts_per_t,
            )

        out_shape = (len(idx), len(self.active_tracks))

        # ragged (b t [p] l)
        tracks = Ragged.from_offsets(out, out_shape, out_offsets)

        return tracks

    def write_transformed_track(
        self,
        new_track: str,
        existing_track: str,
        transform: Callable[
            [NDArray[np.intp], NDArray[np.intp], Ragged[np.float32]],
            Ragged[np.float32],
        ],
        path: Path,
        regions: NDArray[np.int32],
        max_jitter: int,
        idxer: DatasetIndexer,
        haps: Haps | None = None,
        max_mem: int = 2**30,
        overwrite: bool = False,
    ) -> Tracks:
        if new_track == existing_track:
            raise ValueError(
                "New track name must be different from existing track name."
            )

        if existing_track not in self.intervals:
            raise ValueError(
                f"Requested existing track {existing_track} does not exist."
            )

        intervals = self.intervals[existing_track]

        out_dir = path / "intervals" / new_track

        if out_dir.exists() and not overwrite:
            raise FileExistsError(
                f"Track at {out_dir} already exists. Set overwrite=True to overwrite."
            )
        elif out_dir.exists() and overwrite:
            # according to GVL file format, should only have intervals.npy and offsets.npy in here
            for p in out_dir.iterdir():
                p.unlink()
            out_dir.rmdir()

        out_dir.mkdir(parents=True, exist_ok=True)

        # (r)
        n_regions, n_samples = idxer.full_shape
        jittered_regions = regions.copy()
        jittered_regions[:, 1] -= max_jitter
        jittered_regions[:, 2] += max_jitter
        r_idx = np.arange(n_regions)[:, None]
        s_idx = np.arange(n_samples)
        # (r s) -> (r*s)
        ds_idx = np.ravel_multi_index((r_idx, s_idx), idxer.full_shape).ravel()
        r_idx, s_idx = np.unravel_index(ds_idx, idxer.full_shape)
        if haps is not None:
            # extend ends by max hap diff to match write implementation
            jittered_regions[:, 2] += (
                haps._haplotype_ilens(ds_idx, jittered_regions, True)
                .reshape(n_regions, n_samples, haps.genotypes.ploidy)
                .max((1, 2))
                .clip(min=0)
            )
        lengths = jittered_regions[:, 2] - jittered_regions[:, 1]
        # for each region:
        # bytes = (4 bytes / bp) * (bp / sample) * samples
        n_regions, n_samples = intervals.shape
        mem_per_region = 4 * lengths * n_samples
        splits = splits_sum_le_value(mem_per_region, max_mem)
        memmap_intervals_offset = 0
        memmap_offsets_offset = 0
        last_offset = 0
        with tqdm(total=len(splits) - 1) as pbar:
            for offset_s, offset_e in zip(splits[:-1], splits[1:]):
                n_regions = int(offset_e - offset_s)
                ir_idx = repeat(
                    np.arange(offset_s, offset_e, dtype=np.intp),
                    "r -> (r s)",
                    s=n_samples,
                )
                is_idx = repeat(
                    np.arange(n_samples, dtype=np.intp), "s -> (r s)", r=n_regions
                )
                ds_idx = np.ravel_multi_index((ir_idx, is_idx), idxer.full_shape)
                ds_idx = idxer.i2d_map[ds_idx]
                r_idx, s_idx = np.unravel_index(ds_idx, idxer.full_shape)

                pbar.set_description("Writing (decompressing)")
                # (r*s)
                _regions = jittered_regions[r_idx]
                # (r*s+1)
                offsets = _lengths_to_offsets(_regions[:, 2] - _regions[:, 1])
                # layout is (regions, samples) so all samples are local for statistics
                tracks = np.empty(offsets[-1], np.float32)
                intervals_to_tracks(
                    offset_idxs=ds_idx,
                    starts=_regions[:, 1],
                    intervals=intervals.data,
                    itv_offsets=intervals.offsets,
                    out=tracks,
                    out_offsets=offsets,
                )
                tracks = Ragged.from_offsets(tracks, (n_regions, n_samples), offsets)

                pbar.set_description("Writing (transforming)")
                transformed_tracks = transform(ir_idx, is_idx, tracks)
                np.testing.assert_equal(tracks.shape, transformed_tracks.shape)

                pbar.set_description("Writing (compressing)")
                itvs, interval_offsets = tracks_to_intervals(
                    _regions, transformed_tracks.data, transformed_tracks.offsets
                )
                np.testing.assert_equal(
                    len(interval_offsets), n_regions * n_samples + 1
                )

                out = np.memmap(
                    out_dir / "intervals.npy",
                    dtype=itvs.dtype,
                    mode="w+" if memmap_intervals_offset == 0 else "r+",
                    shape=itvs.shape,
                    offset=memmap_intervals_offset,
                )
                out[:] = itvs[:]
                out.flush()
                memmap_intervals_offset += out.nbytes

                interval_offsets += last_offset
                last_offset = interval_offsets[-1]
                out = np.memmap(
                    out_dir / "offsets.npy",
                    dtype=interval_offsets.dtype,
                    mode="w+" if memmap_offsets_offset == 0 else "r+",
                    shape=len(interval_offsets) - 1,
                    offset=memmap_offsets_offset,
                )
                out[:] = interval_offsets[:-1]
                out.flush()
                memmap_offsets_offset += out.nbytes
                pbar.update()

        out = np.memmap(
            out_dir / "offsets.npy",
            dtype=np.int64,
            mode="r+",
            shape=1,
            offset=memmap_offsets_offset,
        )
        out[-1] = last_offset
        out.flush()

        return self.from_path(path, regions, n_samples).with_tracks(self.active_tracks)


@define
class SeqsTracks(Reconstructor[Tuple[Ragged[np.bytes_], Ragged[np.float32]]]):
    seqs: Seqs
    tracks: Tracks

    def __call__(
        self,
        idx: NDArray[np.integer],
        regions: NDArray[np.int32],
        output_length: Union[Literal["ragged", "variable"], int],
        jitter: int,
        rng: Optional[np.random.Generator],
    ) -> Tuple[Ragged[np.bytes_], Ragged[np.float32]]:
        seqs = self.seqs(
            idx=idx,
            regions=regions,
            output_length=output_length,
            jitter=jitter,
            rng=rng,
        )
        tracks = self.tracks(
            idx=idx,
            regions=regions,
            output_length=output_length,
            jitter=jitter,
            rng=rng,
        )
        return seqs, tracks


@define
class HapsTracks(Reconstructor[tuple[H, Ragged[np.float32]]]):
    haps: Haps[H]
    tracks: Tracks

    @overload
    def with_annot(
        self, annotations: Literal[False]
    ) -> HapsTracks[Ragged[np.bytes_]]: ...
    @overload
    def with_annot(
        self, annotations: Literal[True]
    ) -> HapsTracks[RaggedAnnotatedHaps]: ...

    def with_annot(self, annotations: bool) -> HapsTracks:
        haps = self.haps.with_annot(annotations)
        return evolve(self, haps=haps)

    def __call__(
        self,
        idx: NDArray[np.integer],  # (b)
        regions: NDArray[np.int32],  # (b 3)
        output_length: Union[Literal["ragged", "variable"], int],
        jitter: int,
        rng: Optional[np.random.Generator],
    ) -> Tuple[H, Ragged[np.float32]]:
        lengths = regions[:, 2] - regions[:, 1]

        # ragged (b p l), (b p), (b p), (b*p*v), (b*p+1), (b p)
        haps, geno_idx, shifts, diffs, hap_lengths, keep, keep_offsets = (
            self.haps.get_haps_and_shifts(
                idx=idx,
                regions=regions,
                output_length=output_length,
                jitter=jitter,
                rng=rng,
            )
        )

        if isinstance(output_length, int):
            # (b p)
            out_lengths = np.full_like(hap_lengths, output_length + 2 * jitter)
        else:
            # (b p)
            out_lengths = hap_lengths

        # (b) = lengths + max deletion length across ploidy
        track_lengths = lengths - diffs.clip(max=0).min(1)

        # (b*p+1)
        out_ofsts_per_t = _lengths_to_offsets(out_lengths)
        # (b+1)
        track_ofsts_per_t = _lengths_to_offsets(track_lengths)
        n_per_track = out_ofsts_per_t[-1]
        # ragged (b t p l)
        out = np.empty(len(self.tracks.active_tracks) * n_per_track, np.float32)
        out_lens = repeat(out_lengths, "b p -> b t p", t=len(self.tracks.active_tracks))
        out_offsets = _lengths_to_offsets(out_lens)

        for track_ofst, name in enumerate(self.tracks.active_tracks):
            intervals = self.tracks.intervals[name]

            # ragged (b l)
            _tracks = np.empty(track_ofsts_per_t[-1], np.float32)
            intervals_to_tracks(
                starts=regions[:, 1],  # (b)
                offset_idxs=idx,  # (b)
                intervals=intervals.data,  # (r*s*l)
                itv_offsets=intervals.offsets,  # (r*s+1)
                out=_tracks,  # (b*l)
                out_offsets=track_ofsts_per_t,  # (b+1)
            )

            _out = out[track_ofst * n_per_track : (track_ofst + 1) * n_per_track]
            shift_and_realign_tracks_sparse(
                out=_out,  # (b*p*l)
                out_offsets=out_ofsts_per_t,  # (b*p+1)
                regions=regions,  # (b, 3)
                shifts=shifts,  # (b p)
                geno_offset_idxs=geno_idx,  # (b p)
                geno_v_idxs=self.haps.genotypes.variant_idxs,  # (r*s*p*v)
                geno_offsets=self.haps.genotypes.offsets,  # (r*s*p+1)
                positions=self.haps.variants.positions,  # (tot_v)
                sizes=self.haps.variants.sizes,  # (tot_v)
                tracks=_tracks,  # ragged (b l)
                track_offsets=track_ofsts_per_t,  # (b+1)
                keep=keep,  # (b*p*v)
                keep_offsets=keep_offsets,  # (b*p+1)
            )

        out_shape = (
            len(idx),
            len(self.tracks.active_tracks),
            self.haps.genotypes.ploidy,
        )

        # ragged (b t [p] l)
        tracks = Ragged.from_offsets(out, out_shape, out_offsets)

        return haps, tracks



from pathlib import Path

import genvarloader as gvl
import numpy as np
from genvarloader._utils import idx_like_to_array
from pytest import fixture
from pytest_cases import parametrize_with_cases

DATA_DIR = Path(__file__).resolve().parents[1] / "data"
REF = DATA_DIR / "fasta" / "Homo_sapiens.GRCh38.dna.primary_assembly.fa.bgz"
DATASET = gvl.Dataset.open(DATA_DIR / "phased_dataset.gvl", REF)


@fixture(scope="session")
def dataset():
    ds = gvl.Dataset.open(DATA_DIR / "phased_dataset.gvl", REF)
    return ds


def idx_none():
    return


def idx_scalar():
    return 0


def idx_neg_scalar():
    return -1


def idx_slice_none():
    return slice(None)


def idx_slice_start_none():
    return slice(1, None)


def idx_slice_none_stop():
    return slice(None, 2)


def idx_list():
    return [0, 1, 2]


def idx_array():
    return np.arange(3)


def idx_bool():
    return np.random.random(8) > 0.5


def smp_none():
    return


def smp_scalar():
    return 0


def smp_neg_scalar():
    return -1


def smp_slice_none():
    return slice(None)


def smp_slice_start_none():
    return slice(1, None)


def smp_slice_none_stop():
    return slice(None, 2)


def smp_list():
    return [0, 1, 2]


def smp_array():
    return np.arange(3)


def smp_bool():
    return np.random.random(3) > 0.5


def smp_str():
    return DATASET.samples[0]


def smp_strs():
    return DATASET.samples[:2]


@parametrize_with_cases("regions", cases=".", prefix="idx_")
@parametrize_with_cases("samples", cases=".", prefix="smp_")
def test_subset(dataset: gvl.Dataset, regions, samples):
    actual = dataset.subset_to(regions, samples).shape

    if isinstance(regions, int):
        regions = [regions]
    elif regions is None:
        regions = slice(None)

    regions = idx_like_to_array(regions, dataset.n_regions)

    if isinstance(samples, int):
        samples = [samples]
    elif samples is None:
        samples = slice(None)
    elif isinstance(samples, str):
        samples = [dataset.samples.index(samples)]
    elif isinstance(samples, list) and isinstance(samples[0], str):
        samples = [dataset.samples.index(s) for s in samples]

    samples = idx_like_to_array(samples, dataset.n_samples)

    desired = (len(regions), len(samples))
    assert actual == desired



import re
from pathlib import Path
from typing import Union

VCF_SUFFIX = re.compile(r"\.[vb]cf(\.gz)?$")


def path_is_vcf(path: Union[str, Path]) -> bool:
    path = Path(path)
    return VCF_SUFFIX.search(path.name) is not None


def path_is_pgen(path: Union[str, Path]) -> bool:
    path = Path(path)
    return path.suffix == ".pgen" or Path(str(path) + ".pgen").exists()



import random
from typing import Optional, Tuple

import numba as nb
import numpy as np
from attrs import define
from numpy.typing import NDArray

from .._types import ListIdx
from .._utils import _lengths_to_offsets

__all__ = []


@define
class DenseGenotypes:
    """Dense genotypes. In this format, genotypes are stored as a special case of a ragged 3D array where
    each sample has the same number of variants, but each region may have a different number of variants.
    Thus, the first variant indices are the same for every sample, and the offsets are readily computed
    from the first sample's offsets given the number of samples. The genotypes are laid out in C order such
    that the first `n_variant` rows are the genotypes for the first sample, the next `n_variant` rows are
    the genotypes for the second sample, and so on.

    Attributes
    ----------
    genos : NDArray[np.int8]
        Shape = (n_samples * n_variants, ploidy) Genotypes.
    first_v_idxs : NDArray[np.int32]
        Shape = (n_regions,) First variant index for each region.
    offsets : NDArray[np.int64]
        Shape = (n_regions + 1,) Offsets into genos.
    n_samples : int
        Number of samples.
    """

    genos: NDArray[np.int8]  # (n_samples * n_variants, ploidy)
    first_v_idxs: NDArray[np.int32]  # (n_regions)
    offsets: NDArray[np.int64]  # (n_regions + 1)
    n_samples: int

    @property
    def n_regions(self) -> int:
        return len(self.first_v_idxs)

    @property
    def n_variants(self) -> int:
        return len(self.genos) // self.n_samples

    def __len__(self) -> int:
        return len(self.first_v_idxs)

    def __getitem__(self, idx: Tuple[ListIdx, ListIdx]) -> "DenseGenotypes":
        s_idx = idx[0]
        r_idx = idx[1]
        genos = []
        first_v_idxs = self.first_v_idxs[r_idx]
        offsets = np.empty(len(r_idx) + 1, dtype=np.uint32)
        offsets[0] = 0
        shifts = np.asarray(s_idx) * self.n_variants
        for output_idx, (shift, region) in enumerate(zip(shifts, r_idx), 1):
            s, e = self.offsets[region] + shift, self.offsets[region + 1] + shift
            offsets[output_idx] = e - s
            if e > s:
                genos.append(self.genos[s:e])
        if len(genos) == 0:
            genos = np.empty((0, self.genos.shape[1]), dtype=self.genos.dtype)
        else:
            genos = np.concatenate(genos)
        offsets = offsets.cumsum(dtype=np.uint32)

        return DenseGenotypes(genos, first_v_idxs, offsets, self.n_samples)


@nb.njit(parallel=True, nogil=True, cache=True)
def first_v_idxs_to_all_v_idxs(first_variant_indices: NDArray, n_per_region: NDArray):
    """Convert first variant indices to variant indices."""
    out = np.empty(n_per_region.sum(), dtype=np.int32)
    out_start = np.empty_like(n_per_region)
    out_start[0] = 0
    out_start[1:] = n_per_region[:-1].cumsum()
    for i in nb.prange(len(first_variant_indices)):
        f = first_variant_indices[i]
        n = n_per_region[i]
        if n == 0:
            continue
        o_s = out_start[i]
        out[o_s : o_s + n] = np.arange(f, f + n, dtype=np.int32)
    return out


@nb.njit(parallel=True, nogil=True, cache=True)
def get_haplotype_region_ilens(
    genos: NDArray[np.int8],
    first_v_idxs: NDArray[np.int32],
    offsets: NDArray[np.int64],
    ilens: NDArray[np.int32],
):
    n_regions = len(first_v_idxs)
    n_samples = genos.shape[0]
    ploidy = genos.shape[1]
    r_ilens = np.zeros((n_samples, ploidy, n_regions), np.int32)
    for r in nb.prange(n_regions):
        o_s, o_e = offsets[r], offsets[r + 1]
        n_v = o_e - o_s
        if n_v == 0:
            continue
        fvi = first_v_idxs[r]
        for s in nb.prange(n_samples):
            for p in nb.prange(ploidy):
                r_ilens[s, p, r] = np.where(
                    genos[s, p, o_s:o_e] == 1, ilens[fvi : fvi + n_v], 0
                ).sum()
    return r_ilens


@define
class SparseGenotypes:
    """Sparse genotypes corresponding to distinct regions. In this format, genotypes are stored as a ragged 3D array where each
    sample, ploid, and region may have a different number of variants, since unknown and REF genotypes are not stored. The
    variant indices are aligned to the genotypes. Physically, the genotypes and variant indices are stored as 1D arrays.
    Then, each sample and region's info can be sliced out using the offsets:
    >>> i = np.ravel_multi_index((r, p, s), (n_regions, ploidy, n_samples))
    >>> variant_idxs[offsets[i]:offsets[i+1]]

    Attributes
    ----------
    variant_idxs : NDArray[np.int32]
        Shape = (variants * samples * ploidy) Variant indices.
    offsets : NDArray[np.int32]
        Shape = (regions * samples * ploidy + 1) Offsets into genos.
    n_samples : int
        Number of samples.
    ploidy : int
        Ploidy.
    n_regions : int
        Number of regions.
    """

    variant_idxs: NDArray[np.int32]  # (variants * samples * ploidy)
    offsets: NDArray[np.int64]  # (regions * samples * ploidy + 1)
    n_regions: int
    n_samples: int
    ploidy: int
    dosage: Optional[NDArray[np.float32]] = None  # (variants * samples)

    @property
    def effective_shape(self):
        return (self.n_regions, self.n_samples, self.ploidy)

    @classmethod
    def empty(cls, n_regions: int, n_samples: int, ploidy: int):
        """Create an empty sparse genotypes object."""
        return cls(
            np.empty(0, np.int32),
            np.zeros(n_regions * n_samples * ploidy + 1, np.int64),
            n_regions,
            n_samples,
            ploidy,
        )

    @property
    def is_empty(self) -> bool:
        return len(self.variant_idxs) == 0

    def vars(self, region: int, sample: int, ploidy: int):
        """Get variant indices for a given sample and region."""
        i = np.ravel_multi_index(
            (region, sample, ploidy), (self.n_regions, self.n_samples, self.ploidy)
        )
        vars = self.variant_idxs[self.offsets[i] : self.offsets[i + 1]]
        return vars

    def concat(*genos: "SparseGenotypes") -> "SparseGenotypes":
        """Concatenate sparse genotypes."""

        if not all(g.n_samples == genos[0].n_samples for g in genos):
            raise ValueError("All genotypes must have the same number of samples.")
        if not all(g.ploidy == genos[0].ploidy for g in genos):
            raise ValueError("All genotypes must have the same ploidy.")

        total_n_regions = sum(g.n_regions for g in genos)
        variant_idxs = np.concatenate([g.variant_idxs for g in genos])
        offsets = _lengths_to_offsets(
            np.concatenate([np.diff(g.offsets) for g in genos])
        )
        return SparseGenotypes(
            variant_idxs=variant_idxs,
            offsets=offsets,
            n_regions=total_n_regions,
            n_samples=genos[0].n_samples,
            ploidy=genos[0].ploidy,
        )

    @classmethod
    def from_dense(
        cls,
        genos: NDArray[np.int8],
        first_v_idxs: NDArray[np.int32],
        offsets: NDArray[np.int64],
        dosages: Optional[NDArray[np.float32]] = None,
    ):
        """Convert dense genotypes to sparse genotypes.

        Parameters
        ----------
        genos : NDArray[np.int8]
            Shape = (sample, ploidy, variants) Genotypes.
        first_v_idxs : NDArray[np.uint32]
            Shape = (regions) First variant index for each region.
        offsets : NDArray[np.uint32]
            Shape = (regions + 1) Offsets into genos.
        dosages : Optional[NDArray[np.float32]]
            Shape = (sample, ploidy, variants) Dosages.
        """
        n_regions = len(first_v_idxs)
        n_samples = genos.shape[0]
        ploidy = genos.shape[1]
        # (s p v)
        keep = genos == 1
        n_per_rsp = get_n_per_rsp(keep, offsets, n_regions)
        sparse_offsets = _lengths_to_offsets(n_per_rsp.ravel(), np.int64)
        variant_idxs = keep_mask_to_rsp_v_idx(
            keep, first_v_idxs, offsets, sparse_offsets, n_regions, n_samples, ploidy
        )
        if dosages is not None:
            dosages = dosages[keep]
        return cls(
            variant_idxs=variant_idxs,
            offsets=sparse_offsets,
            n_regions=n_regions,
            n_samples=n_samples,
            ploidy=ploidy,
            dosage=dosages,
        )

    @classmethod
    def from_dense_with_length(
        cls,
        genos: NDArray[np.int8],
        first_v_idxs: NDArray[np.int32],
        offsets: NDArray[np.int64],
        ilens: NDArray[np.int32],
        positions: NDArray[np.int32],
        starts: NDArray[np.int32],
        lengths: NDArray[np.int32],
    ):
        """Convert dense genotypes to sparse genotypes.

        Parameters
        ----------
        genos : NDArray[np.int8]
            Shape = (sample, ploidy, variants) Genotypes.
        first_v_idxs : NDArray[np.uint32]
            Shape = (regions) First variant index for each region.
        offsets : NDArray[np.uint32]
            Shape = (regions + 1) Offsets into genos.
        ilens : NDArray[np.int32]
            Shape = (total_variants) ILEN of all unique variants.
        positions : NDArray[np.int32]
            Shape = (total_variants) Positions of unique variants.
        starts : NDArray[np.int32]
            Shape = (regions) Start of query regions.
        lengths : NDArray[np.int32]
            Shape = (regions) Lengths of the output haplotypes.
        """
        n_regions = len(first_v_idxs)
        n_samples = genos.shape[0]
        ploidy = genos.shape[1]
        # (s p v)
        keep, min_ilens = get_keep_mask_for_length(
            genos,
            offsets,
            first_v_idxs,
            positions,
            ilens,
            starts,
            lengths,
        )
        # (r)
        max_ends: NDArray[np.int32] = starts + lengths - min_ilens.clip(max=0)
        # (r s p)
        n_per_rsp = get_n_per_rsp(keep, offsets, n_regions)
        sparse_offsets = _lengths_to_offsets(n_per_rsp.ravel(), np.int64)
        variant_idxs = keep_mask_to_rsp_v_idx(
            keep, first_v_idxs, offsets, sparse_offsets, n_regions, n_samples, ploidy
        )
        sparse_genos = cls(
            variant_idxs=variant_idxs,
            offsets=sparse_offsets,
            n_regions=n_regions,
            n_samples=n_samples,
            ploidy=ploidy,
        )
        return sparse_genos, max_ends


@nb.njit(parallel=True, nogil=True, cache=True)
def get_keep_mask_for_length(
    genos: NDArray[np.int8],
    offsets: NDArray[np.int64],
    first_v_idxs: NDArray[np.int32],
    positions: NDArray[np.int32],
    ilens: NDArray[np.int32],
    starts: NDArray[np.int32],
    lengths: NDArray[np.int32],
):
    """Mark genotypes to keep based on being an ALT allele and being within the length of the haplotype.

    Parameters
    ----------
    genos : NDArray[np.int8]
        Shape = (samples, ploidy, variants) Genotypes.
    offsets : NDArray[np.int32]
        Shape = (regions + 1) Offsets into genos.
    first_v_idxs : NDArray[np.int32]
        Shape = (regions) First variant index for each region.
    positions : NDArray[np.int32]
        Shape = (total_variants) Positions of variants.
    ilens : NDArray[np.int32]
        Shape = (total_variants) ILEN of all unique variants.
    starts : NDArray[np.int32]
        Shape = (regions) Start of query regions.
    lengths : NDArray[np.int32]
        Shape = (regions) Length of haplotypes.
    """
    n_samples = genos.shape[0]
    ploidy = genos.shape[1]
    n_regions = len(starts)
    keep = np.empty_like(genos, np.bool_)
    out_ilens = np.zeros((n_regions, n_samples, ploidy), np.int32)
    for r in nb.prange(n_regions):
        v_s, v_e = offsets[r], offsets[r + 1]
        n_variants = v_e - v_s
        if n_variants == 0:
            continue
        ref_start = starts[r]
        length = lengths[r]
        for s in nb.prange(n_samples):
            for p in range(ploidy):
                cum_ilen = 0
                rel_ref_end = 0
                cum_write_len = 0
                for rel_v_idx in range(v_s, v_e):
                    if genos[s, p, rel_v_idx] == 1:
                        abs_v_idx = first_v_idxs[r] + rel_v_idx - v_s
                        rel_pos = positions[abs_v_idx] - ref_start
                        maybe_add_one = (
                            rel_pos >= 0
                        )  # length of alt allele is +1 only if rel_pos >= 0
                        ilen: int = ilens[abs_v_idx]  # type: ignore

                        # add dist from last variant to current variant
                        cum_write_len += rel_pos - rel_ref_end

                        # do we need this variant to reach the length?
                        if cum_write_len < length:
                            keep[s, p, rel_v_idx] = True

                            # update rel_ref_end to end of variant
                            v_rel_end = rel_pos - min(0, ilen) + maybe_add_one
                            rel_ref_end = v_rel_end + maybe_add_one

                            # update cum_write_len and cum_ilen
                            missing_len = length - cum_write_len
                            v_len = (
                                max(rel_pos, 0)
                                - rel_ref_end
                                + max(0, ilen)
                                + maybe_add_one
                            )
                            clip_right = max(0, v_len - missing_len)
                            v_len -= clip_right
                            cum_write_len += v_len
                            cum_ilen += ilen - clip_right
                    else:
                        keep[s, p, rel_v_idx] = False
                out_ilens[r, s, p] = cum_ilen

    for r in nb.prange(n_regions):
        for s in nb.prange(n_samples):
            out_ilens[r, s, 0] = out_ilens[r, s, :].min()
        out_ilens[r, 0, 0] = out_ilens[r, :, 0].min()
    min_ilens = out_ilens[:, 0, 0]
    return keep, min_ilens


@nb.njit(parallel=True, nogil=True, cache=True)
def get_n_per_rsp(keep: NDArray[np.bool_], offsets: NDArray[np.int64], n_regions: int):
    n_samples, ploidy, _ = keep.shape
    n_per_rsp = np.empty((n_regions, n_samples, ploidy), np.int32)
    for r in nb.prange(n_regions):
        o_s, o_e = offsets[r], offsets[r + 1]
        for s in nb.prange(n_samples):
            for p in nb.prange(ploidy):
                n_per_rsp[r, s, p] = keep[s, p, o_s:o_e].sum()
    return n_per_rsp


@nb.njit(parallel=True, nogil=True, cache=True)
def keep_mask_to_rsp_v_idx(
    keep: NDArray[np.bool_],  # (s p v)
    first_v_idxs: NDArray[np.int32],  # (r)
    offsets: NDArray[np.int64],  # (r + 1)
    sparse_offsets: NDArray[np.int64],  # (r*s*p + 1)
    n_regions,
    n_samples,
    ploidy,
):
    variant_idxs = np.empty(sparse_offsets[-1], np.int32)
    for r in nb.prange(n_regions):
        fvi = first_v_idxs[r]
        o_s, o_e = offsets[r], offsets[r + 1]
        n_variants = o_e - o_s
        if n_variants == 0:
            continue
        for s in nb.prange(n_samples):
            for p in nb.prange(ploidy):
                out_start = sparse_offsets[r * n_samples * ploidy + s * ploidy + p]
                out_step = 0
                for v in range(n_variants):
                    if keep[s, p, o_s + v]:
                        variant_idxs[out_start + out_step] = fvi + v
                        out_step += 1
    return variant_idxs


@define
class SparseSomaticGenotypes:
    """Sparse, unphased, somatic genotypes corresponding to distinct regions. In this format, genotypes are stored as a ragged 2D array where each
    sample and region may have a different number of variants, since unknown and REF genotypes are not stored. The
    variant indices are aligned to the genotypes. Physically, the genotypes and variant indices are stored as 1D arrays.
    Then, each sample and region's info can be sliced out using the offsets:
    >>> i = np.ravel_multi_index((r, s), (n_regions, n_samples))
    >>> variant_idxs[offsets[i]:offsets[i+1]]
    >>> ccfs[offsets[i]:offsets[i+1]]

    Attributes
    ----------
    variant_idxs : NDArray[np.int32]
        Shape = (variants * samples) Variant indices.
    ccfs : NDArray[np.float32]
        Shape = (variants * samples) Cancer cell fractions (CCF).
    offsets : NDArray[np.int32]
        Shape = (regions * samples * ploidy + 1) Offsets into genos.
    n_samples : int
        Number of samples.
    n_regions : int
        Number of regions.
    """

    variant_idxs: NDArray[np.int32]  # (variants * samples)
    ccfs: NDArray[np.float32]  # (variants * samples)
    offsets: NDArray[np.int64]  # (regions * samples + 1)
    n_regions: int
    n_samples: int
    ploidy = 1

    @property
    def effective_shape(self):
        """Effective shape of the sparse genotypes (n_regions, n_samples, ploidy)
        where ploidy is always represented as 1. The ploidy is treated as 1 to be consistent with."""
        return (self.n_regions, self.n_samples, self.ploidy)

    @classmethod
    def empty(cls, n_regions: int, n_samples: int):
        """Create an empty sparse genotypes object."""
        return cls(
            np.empty(0, np.int32),
            np.empty(0, np.float32),
            np.zeros(n_regions * n_samples + 1, np.int64),
            n_regions,
            n_samples,
        )

    @property
    def is_empty(self) -> bool:
        return len(self.variant_idxs) == 0

    def vars(self, region: int, sample: int):
        """Get variant indices for a given sample and region."""
        i = np.ravel_multi_index((region, sample), (self.n_regions, self.n_samples))
        vars = self.variant_idxs[self.offsets[i] : self.offsets[i + 1]]
        return vars

    def concat(*genos: "SparseSomaticGenotypes") -> "SparseSomaticGenotypes":
        """Concatenate sparse genotypes."""

        if not all(g.n_samples == genos[0].n_samples for g in genos):
            raise ValueError("All genotypes must have the same number of samples.")

        total_n_regions = sum(g.n_regions for g in genos)
        variant_idxs = np.concatenate([g.variant_idxs for g in genos])
        offsets = _lengths_to_offsets(
            np.concatenate([np.diff(g.offsets) for g in genos])
        )

        ccfs = np.concatenate([g.ccfs for g in genos if g.ccfs is not None])

        return SparseSomaticGenotypes(
            variant_idxs=variant_idxs,
            offsets=offsets,
            n_regions=total_n_regions,
            n_samples=genos[0].n_samples,
            ccfs=ccfs,
        )

    @classmethod
    def from_dense(
        cls,
        genos: NDArray[np.int8],
        first_v_idxs: NDArray[np.int32],
        offsets: NDArray[np.int64],
        ccfs: NDArray[np.float32],
    ):
        """Convert dense genotypes to sparse genotypes. Genotypes will be considered ALT if any allele is ALT.
        e.g. 0/1 -> 1, 1/1 -> 1, 0/0 -> 0, 0/2 -> 1.

        Parameters
        ----------
        genos : NDArray[np.int8]
            Shape = (sample, ploidy, variants) Genotypes.
        first_v_idxs : NDArray[np.uint32]
            Shape = (regions) First variant index for each region.
        offsets : NDArray[np.uint32]
            Shape = (regions + 1) Offsets into genos.
        ccfs : Optional[NDArray[np.float32]]
            Shape = (sample, variants) Cancer cell fractions (CCF).
        """
        n_regions = len(first_v_idxs)
        n_samples = genos.shape[0]
        # (s p v) -> (s v)
        keep = (genos == 1).any(1)
        # (r s)
        n_per_rs = get_n_per_rs(keep, offsets, n_regions)
        sparse_offsets = _lengths_to_offsets(n_per_rs.ravel(), np.int64)
        variant_idxs = keep_mask_to_rs_v_idx(
            keep, first_v_idxs, offsets, sparse_offsets, n_regions, n_samples
        )
        # (s v) -> region/variant-major, flattened
        ccfs = ccfs.T[keep.T]
        return cls(
            variant_idxs=variant_idxs,
            ccfs=ccfs,
            offsets=sparse_offsets,
            n_regions=n_regions,
            n_samples=n_samples,
        )

    @classmethod
    def from_dense_with_length(
        cls,
        genos: NDArray[np.int8],
        first_v_idxs: NDArray[np.int32],
        offsets: NDArray[np.int64],
        ilens: NDArray[np.int32],
        positions: NDArray[np.int32],
        starts: NDArray[np.int32],
        lengths: NDArray[np.int32],
        ccfs: NDArray[np.float32],
    ):
        """Convert dense genotypes to sparse genotypes.

        Parameters
        ----------
        genos : NDArray[np.int8]
            Shape = (sample, ploidy, variants) Genotypes.
        first_v_idxs : NDArray[np.uint32]
            Shape = (regions) First variant index for each region.
        offsets : NDArray[np.uint32]
            Shape = (regions + 1) Offsets into genos.
        ilens : NDArray[np.int32]
            Shape = (total_variants) ILEN of all unique variants.
        positions : NDArray[np.int32]
            Shape = (total_variants) Positions of unique variants.
        starts : NDArray[np.int32]
            Shape = (regions) Start of query regions.
        lengths : NDArray[np.int32]
            Shape = (regions) Desired lengths of the output haplotypes.
        ccfs : NDArray[np.float32]
            Shape = (sample, variants) Cancer cell fractions (CCF).
        """
        n_regions = len(first_v_idxs)
        n_samples = genos.shape[0]
        # (s p v)
        keep, min_ilens = get_keep_mask_for_length(
            genos,
            offsets,
            first_v_idxs,
            positions,
            ilens,
            starts,
            lengths,
        )
        keep = keep.any(1)
        # (r)
        max_ends: NDArray[np.int32] = starts + lengths - min_ilens.clip(max=0)
        # (r s)
        n_per_rs = get_n_per_rs(keep, offsets, n_regions)
        sparse_offsets = _lengths_to_offsets(n_per_rs.ravel(), np.int64)
        variant_idxs = keep_mask_to_rs_v_idx(
            keep, first_v_idxs, offsets, sparse_offsets, n_regions, n_samples
        )
        # (s v) -> region/variant-major, flattened
        ccfs = ccfs.T[keep.T]
        sparse_genos = cls(
            variant_idxs=variant_idxs,
            ccfs=ccfs,
            offsets=sparse_offsets,
            n_regions=n_regions,
            n_samples=n_samples,
        )
        return sparse_genos, max_ends


@nb.njit(parallel=True, nogil=True, cache=True)
def get_n_per_rs(keep, offsets, n_regions):
    n_samples, n_variants = keep.shape
    n_per_rs = np.empty((n_regions, n_samples), np.int32)
    for r in nb.prange(n_regions):
        o_s, o_e = offsets[r], offsets[r + 1]
        for s in nb.prange(n_samples):
            n_per_rs[r, s] = keep[s, o_s:o_e].sum()
    return n_per_rs


@nb.njit(parallel=True, nogil=True, cache=True)
def keep_mask_to_rs_v_idx(
    keep,  # (s v)
    first_v_idxs,  # (r)
    offsets,  # (r + 1)
    sparse_offsets,  # (r*s + 1)
    n_regions,
    n_samples,
):
    variant_idxs = np.empty(sparse_offsets[-1], np.int32)
    for r in nb.prange(n_regions):
        fvi = first_v_idxs[r]
        o_s, o_e = offsets[r], offsets[r + 1]
        n_variants = o_e - o_s
        if n_variants == 0:
            continue
        for s in nb.prange(n_samples):
            out_start = sparse_offsets[r * n_samples + s]
            out_step = 0
            for v in range(n_variants):
                if keep[s, o_s + v]:
                    variant_idxs[out_start + out_step] = fvi + v
                    out_step += 1
    return variant_idxs


@nb.njit(parallel=True, nogil=True, cache=True)
def get_diffs(
    first_v_idxs: NDArray[np.uint32],
    offsets: NDArray[np.uint32],
    genotypes: NDArray[np.int8],
    size_diffs: NDArray[np.int32],
) -> NDArray[np.uint32]:
    """Get difference in length wrt reference genome for given genotypes.

    Parameters
    ----------
    first_v_idxs : NDArray[np.uint32]
        Shape = (n_regions,) First variant index for each query.
    offsets : NDArray[np.uint32]
        Shape = (n_regions + 1,) Offsets into genos.
    genotypes : NDArray[np.int8]
        Shape = (n_variants, ploidy) Genotypes.
    size_diffs : NDArray[np.int32]
        Shape = (total_variants,) Size of variants.
    """
    n_regions = len(first_v_idxs)
    ploidy = genotypes.shape[1]
    diffs = np.empty((n_regions, ploidy), np.uint32)

    for region in nb.prange(n_regions):
        o_s, o_e = offsets[region], offsets[region + 1]
        n_variants = o_e - o_s

        if n_variants == 0:
            diffs[region] = 0
            continue

        v_s = first_v_idxs[region]
        v_e = v_s + n_variants
        # (v p)
        genos = genotypes[o_s:o_e]
        # (v p) -> (p)
        diff = np.where(genos == 1, size_diffs[v_s:v_e, None], 0).sum(0).clip(0)
        diffs[region] = diff
    return diffs


@nb.njit(parallel=True, nogil=True, cache=True)
def get_diffs_sparse(
    geno_offset_idxs: NDArray[np.intp],
    geno_v_idxs: NDArray[np.int32],
    geno_offsets: NDArray[np.int64],
    size_diffs: NDArray[np.int32],
    keep: Optional[NDArray[np.bool_]] = None,
    keep_offsets: Optional[NDArray[np.int64]] = None,
    starts: Optional[NDArray[np.int32]] = None,
    ends: Optional[NDArray[np.int32]] = None,
    positions: Optional[NDArray[np.int32]] = None,
):
    """Get difference in length wrt reference genome for given genotypes.

    If starts, ends, & positions are given, they take priority over keep and keep_offsets.

    Parameters
    ----------
    geno_offset_idxs : NDArray[np.intp]
        Shape = (n_regions, ploidy) Indices for each region into offsets.
    geno_v_idxs : NDArray[np.int32]
        Shape = (variants*samples*ploidy) Sparse genotypes i.e. variant indices for ALT genotypes.
    geno_offsets : NDArray[np.int32]
        Shape = (regions*samples*ploidy + 1) Offsets into sparse genotypes.
    size_diffs : NDArray[np.int32]
        Shape = (total_variants) Size of all unique variants.
    keep : Optional[NDArray[np.bool_]]
        Shape = (variants*samples*ploidy) Keep mask for genotypes.
    keep_offsets : Optional[NDArray[np.int64]]
        Shape = (regions*samples*ploidy + 1) Offsets into keep.
    starts : Optional[NDArray[np.int32]]
        Shape = (regions) Start of query regions.
    ends : Optional[NDArray[np.int32]]
        Shape = (regions) End of query regions.
    positions : Optional[NDArray[np.int32]]
        Shape = (total_variants) Positions of unique variants.
    """
    n_queries, ploidy = geno_offset_idxs.shape
    diffs = np.empty((n_queries, ploidy), np.int32)
    for query in nb.prange(n_queries):
        for hap in nb.prange(ploidy):
            o_idx = geno_offset_idxs[query, hap]
            o_s, o_e = geno_offsets[o_idx], geno_offsets[o_idx + 1]
            n_variants = o_e - o_s
            if n_variants == 0:
                diffs[query, hap] = 0
            elif starts is not None and ends is not None and positions is not None:
                diffs[query, hap] = 0
                for v in range(o_s, o_e):
                    if keep is not None and keep_offsets is not None:
                        k_s = keep_offsets[query * ploidy + hap]
                        v_keep = keep[k_s + (v - o_s)]
                        if not v_keep:
                            continue

                    v_idx: int = geno_v_idxs[v]
                    v_pos = positions[v_idx]
                    v_diff = size_diffs[v_idx]
                    # +1 assumes atomized variants
                    v_end = v_pos - min(0, v_diff) + 1

                    if v_end <= starts[query]:
                        # variant doesn't span region
                        continue

                    if v_pos >= ends[query]:
                        # variants are sorted by position so this variant and everything
                        # after will be outside the region
                        break

                    # deletion may start before region
                    #     0 1 2 3 4 5 6
                    # DEL s - - r e - - : +max(0, 3 - 0) -> -3 + 3 = 0
                    # DEL r - s - e - - : +max(0, 0 - 2) -> -1 + 0 = -1
                    # where r is region start, s is variant start, e is variant end (exclusive)
                    # count the "-" to get ilen
                    if v_diff < 0:
                        v_diff += max(0, starts[query] - v_pos)
                    # deletion may end after region
                    v_diff += max(0, v_end - ends[query])

                    diffs[query, hap] += v_diff
            elif keep is not None and keep_offsets is not None:
                v_idxs = geno_v_idxs[o_s:o_e]
                k_idx = query * ploidy + hap
                qh_keep = keep[keep_offsets[k_idx] : keep_offsets[k_idx + 1]]
                v_idxs = v_idxs[qh_keep]
                diffs[query, hap] = size_diffs[v_idxs].sum()
            else:
                diffs[query, hap] = size_diffs[geno_v_idxs[o_s:o_e]].sum()
    return diffs


@nb.njit(parallel=True, nogil=True, cache=True)
def reconstruct_haplotypes_from_sparse(
    out: NDArray[np.uint8],
    out_offsets: NDArray[np.int64],
    regions: NDArray[np.int32],
    shifts: NDArray[np.int32],
    geno_offset_idxs: NDArray[np.intp],
    geno_offsets: NDArray[np.int64],
    geno_v_idxs: NDArray[np.int32],
    positions: NDArray[np.int32],
    sizes: NDArray[np.int32],
    alt_alleles: NDArray[np.uint8],
    alt_offsets: NDArray[np.int64],
    ref: NDArray[np.uint8],
    ref_offsets: NDArray[np.uint64],
    pad_char: int,
    keep: Optional[NDArray[np.bool_]] = None,
    keep_offsets: Optional[NDArray[np.int64]] = None,
    annot_v_idxs: Optional[NDArray[np.int32]] = None,
    annot_ref_pos: Optional[NDArray[np.int32]] = None,
):
    """Reconstruct haplotypes from reference sequence and variants.

    Parameters
    ----------
    offset_idxs: NDArray[np.intp]
        Shape = (n_regions, ploidy) Indices for each region into offsets.
    out : NDArray[np.uint8]
        Ragged array of shape = (n_regions, ploidy) to write haplotypes into.
    out_offsets : NDArray[np.int64]
        Shape = (n_regions*ploidy + 1) Offsets into out.
    regions : NDArray[np.int32]
        Shape = (n_regions, 3) Regions to reconstruct haplotypes.
    shifts : NDArray[np.uint32]
        Shape = (n_regions, ploidy) Shifts for each region.
    offsets : NDArray[np.uint32]
        Shape = (ploidy*n_regions + 1) Offsets into genos.
    sparse_genos : NDArray[np.int32]
        Shape = (variants) Sparse genotypes of variants i.e. variant indices for ALT genotypes.
    positions : NDArray[np.int32]
        Shape = (total_variants) Positions of variants.
    sizes : NDArray[np.int32]
        Shape = (total_variants) Sizes of variants.
    alt_alleles : NDArray[np.uint8]
        Shape = (total_alt_length) ALT alleles.
    alt_offsets : NDArray[np.uintp]
        Shape = (total_variants) Offsets of ALT alleles.
    ref : NDArray[np.uint8]
        Shape = (ref_length) Reference sequence.
    ref_offsets : NDArray[np.uint64]
        Shape = (n_contigs) Offsets of reference sequences.
    pad_char : int
        Padding character.
    keep : Optional[NDArray[np.bool_]]
        Shape = (variants) Keep mask for genotypes.
    annot_v_idxs : Optional[NDArray[np.int32]]
        Ragged array of shape (n_regions, ploidy). Variant indices for annotations.
    annot_ref_pos : Optional[NDArray[np.int32]]
        Ragged array of shape (n_regions, ploidy). Reference positions for annotations.
    """
    n_regions, ploidy = geno_offset_idxs.shape
    for query in nb.prange(n_regions):
        q = regions[query]
        c_idx: int = q[0]
        c_s = ref_offsets[c_idx]
        c_e = ref_offsets[c_idx + 1]
        ref_start: int = q[1]
        _reference = ref[c_s:c_e]

        for hap in nb.prange(ploidy):
            # index for full sparse genos
            o_idx = geno_offset_idxs[query, hap]

            # local index for subset of variants that are implied by offset_idxs
            k_idx = query * ploidy + hap
            if keep is not None and keep_offsets is not None:
                qh_keep = keep[keep_offsets[k_idx] : keep_offsets[k_idx + 1]]
            else:
                qh_keep = None

            # aligned to out sequence
            out_s, out_e = out_offsets[k_idx], out_offsets[k_idx + 1]
            qh_out = out[out_s:out_e]
            qh_shift = shifts[query, hap]

            qh_annot_v_idxs = (
                annot_v_idxs[out_s:out_e] if annot_v_idxs is not None else None
            )
            qh_annot_ref_pos = (
                annot_ref_pos[out_s:out_e] if annot_ref_pos is not None else None
            )

            reconstruct_haplotype_from_sparse(
                offset_idx=o_idx,
                geno_v_idxs=geno_v_idxs,
                geno_offsets=geno_offsets,
                positions=positions,
                sizes=sizes,
                shift=qh_shift,
                alt_alleles=alt_alleles,
                alt_offsets=alt_offsets,
                ref=_reference,
                ref_start=ref_start,
                out=qh_out,
                pad_char=pad_char,
                keep=qh_keep,
                annot_v_idxs=qh_annot_v_idxs,
                annot_ref_pos=qh_annot_ref_pos,
            )


@nb.njit(nogil=True, cache=True)
def reconstruct_haplotype_from_sparse(
    offset_idx: int,
    geno_v_idxs: NDArray[np.int32],
    geno_offsets: NDArray[np.int64],
    positions: NDArray[np.int32],
    sizes: NDArray[np.int32],
    shift: int,
    alt_alleles: NDArray[np.uint8],  # full set
    alt_offsets: NDArray[np.int64],  # full set
    ref: NDArray[np.uint8],  # full contig
    ref_start: int,  # may be negative
    out: NDArray[np.uint8],
    pad_char: int,
    keep: Optional[NDArray[np.bool_]] = None,
    annot_v_idxs: Optional[NDArray[np.int32]] = None,
    annot_ref_pos: Optional[NDArray[np.int32]] = None,
):
    """Reconstruct a haplotype from reference sequence and variants.

    Parameters
    ----------
    offset_idx : int
        Index for `offsets` for where to find the offsets into variant_idxs.
    variant_idxs : int
        Index of alt variants for all samples and variants.
    offsets : NDArray[np.int32]
        Shape = Offsets into variant indices.
    positions : NDArray[np.int32]
        Shape = (total_variants) Positions of variants.
    sizes : NDArray[np.int32]
        Shape = (total_variants) Sizes of variants.
    shift : int
        Total amount to shift by.
    alt_alleles : NDArray[np.uint8]
        Shape = (total_alt_length) ALT alleles.
    alt_offsets : NDArray[np.uintp]
        Shape = (total_variants + 1) Offsets of ALT alleles.
    ref : NDArray[np.uint8]
        Shape = (ref_length) Reference sequence for the whole contig. ref_length >= out_length
    ref_start : int
        Start position of reference sequence, may be negative.
    out : NDArray[np.uint8]
        Shape = (out_length) Output array.
    pad_char : int
        Padding character.
    keep: Optional[NDArray[np.bool_]]
        Shape = (variants) Keep mask for genotypes.
    annot_v_idxs: Optional[NDArray[np.int32]]
        Shape = (out_length) Variant indices for annotations.
    annot_ref_pos: Optional[NDArray[np.int32]]
        Shape = (out_length) Reference positions for annotations
    """
    _variant_idxs = geno_v_idxs[geno_offsets[offset_idx] : geno_offsets[offset_idx + 1]]
    length = len(out)
    n_variants = len(_variant_idxs)

    # where to get next reference subsequence
    ref_idx = ref_start
    # where to put next subsequence
    out_idx = 0
    # how much we've shifted
    shifted = 0

    # if ref_idx is negative, we need to pad the beginning of the haplotype
    if ref_idx < 0:
        pad_len = -ref_idx
        shifted = min(shift, pad_len)
        pad_len -= shifted
        out[out_idx : out_idx + pad_len] = pad_char
        if annot_v_idxs is not None:
            annot_v_idxs[out_idx : out_idx + pad_len] = -1
        if annot_ref_pos is not None:
            annot_ref_pos[out_idx : out_idx + pad_len] = -1
        out_idx += pad_len
        ref_idx = 0

    for v in range(n_variants):
        if keep is not None and not keep[v]:
            continue

        variant: np.int32 = _variant_idxs[v]
        v_pos = positions[variant]
        v_diff = sizes[variant]
        allele = alt_alleles[alt_offsets[variant] : alt_offsets[variant + 1]]
        v_len = len(allele)
        # +1 assumes atomized variants, exactly 1 nt shared between REF and ALT
        v_ref_end = v_pos - min(0, v_diff) + 1

        # if variant is a DEL spanning start of query
        if v_pos < ref_start and v_diff < 0 and v_ref_end >= ref_start:
            ref_idx = v_ref_end
            continue

        # overlapping variants
        # v_rel_pos < ref_idx only if we see an ALT at a given position a second
        # time or more. We'll do what bcftools consensus does and only use the
        # first ALT variant we find.
        if v_pos < ref_idx:
            continue

        # handle shift
        if shifted < shift:
            ref_shift_dist = v_pos - ref_idx
            # not enough distance to finish the shift even with the variant
            if shifted + ref_shift_dist + v_len < shift:
                # skip the variant
                continue
            # enough distance between ref_idx and start of variant to finish shift
            elif shifted + ref_shift_dist >= shift:
                ref_idx += shift - shifted
                shifted = shift
                # can still use the variant and whatever ref is left between
                # ref_idx and the variant
            # ref + all or some of variant is enough to finish shift
            else:
                # how much left to shift - amount of ref we can use
                allele_start_idx = shift - shifted - ref_shift_dist
                shifted = shift
                #! without if statement, parallel=True can cause a SystemError!
                # * parallel jit cannot handle changes in array dimension.
                # * without this, allele can change from a 1D array to a 0D
                # * array.
                # enough dist with variant to complete shift
                if allele_start_idx == v_len:
                    # move ref to end of variant
                    ref_idx = v_ref_end
                    # skip the variant
                    continue
                # consume ref up to beginning of variant
                # ref_idx will be moved to end of variant after using the variant
                ref_idx = v_pos
                # adjust variant to start at allele_start_idx
                allele = allele[allele_start_idx:]
                v_len = len(allele)

        # add reference sequence
        ref_len = v_pos - ref_idx
        if out_idx + ref_len >= length:
            # ref will get written by final clause
            # handles case where extraneous variants downstream of the haplotype were provided
            break
        out[out_idx : out_idx + ref_len] = ref[ref_idx : ref_idx + ref_len]
        if annot_v_idxs is not None:
            annot_v_idxs[out_idx : out_idx + ref_len] = -1
        if annot_ref_pos is not None:
            annot_ref_pos[out_idx : out_idx + ref_len] = np.arange(
                ref_idx, ref_idx + ref_len
            )
        out_idx += ref_len

        # apply variant
        writable_length = min(v_len, length - out_idx)
        out[out_idx : out_idx + writable_length] = allele[:writable_length]
        if annot_v_idxs is not None:
            annot_v_idxs[out_idx : out_idx + writable_length] = variant
        if annot_ref_pos is not None:
            annot_ref_pos[out_idx : out_idx + writable_length] = v_pos
        out_idx += writable_length

        # advance ref_idx to end of variant
        ref_idx = v_ref_end

        if out_idx >= length:
            break

    if shifted < shift:
        # need to shift the rest of the track
        ref_idx += shift - shifted
        ref_idx = min(ref_idx, len(ref))
        shifted = shift

    # fill rest with reference sequence and pad with Ns
    unfilled_length = length - out_idx
    if unfilled_length > 0:
        # fill with reference sequence
        writable_ref = min(unfilled_length, len(ref) - ref_idx)
        out_end_idx = out_idx + writable_ref
        ref_end_idx = ref_idx + writable_ref
        out[out_idx:out_end_idx] = ref[ref_idx:ref_end_idx]
        if annot_v_idxs is not None:
            annot_v_idxs[out_idx:out_end_idx] = -1
        if annot_ref_pos is not None:
            annot_ref_pos[out_idx:out_end_idx] = np.arange(ref_idx, ref_end_idx)

        # pad
        if out_end_idx < length:
            out[out_end_idx:] = pad_char
            if annot_v_idxs is not None:
                annot_v_idxs[out_end_idx:] = -1
            if annot_ref_pos is not None:
                annot_ref_pos[out_end_idx:] = -1


UNSEEN_VARIANT = np.iinfo(np.uint32).max


@nb.njit(parallel=True, nogil=True, cache=True)
def choose_unphased_variants(
    starts: NDArray[np.int32],
    ends: NDArray[np.int32],
    geno_offset_idxs: NDArray[np.intp],
    geno_v_idxs: NDArray[np.int32],
    geno_offsets: NDArray[np.int64],
    positions: NDArray[np.int32],
    sizes: NDArray[np.int32],
    ccfs: NDArray[np.float32],
    deterministic: bool,
) -> Tuple[NDArray[np.bool_], NDArray[np.int64]]:
    """Mark variants to keep for each haplotype.

    Parameters
    ----------
    geno_offset_idxs : NDArray[np.intp]
        Shape = (n_regions, ploidy) Indices for each region into offsets.
    starts : NDArray[np.int32]
        Shape = (n_regions) Start positions for each region.
    offsets : NDArray[np.int64]
        Shape = (total_variants + 1) Offsets into sparse genotypes.
    sparse_genos : NDArray[np.int32]
        Shape = (total_variants) Sparse genotypes i.e. variant indices for ALT genotypes.
    positions : NDArray[np.int32]
        Shape = (total_variants) Positions of variants.
    sizes : NDArray[np.int32]
        Shape = (total_variants) Sizes of variants.
    ccfs : NDArray[np.float32]
        Shape = (total_variants) Cancer cell fractions (CCF) of variants.
    ends : NDArray[np.int32]
        Shape = (n_regions) Ends for each region.
    deterministic : bool
        Whether to deterministically assign variants to groups
    """
    n_regions, ploidy = geno_offset_idxs.shape

    lengths = np.empty((n_regions, ploidy), np.int64)
    for query in nb.prange(n_regions):
        for hap in range(ploidy):
            o_idx = geno_offset_idxs[query, hap]
            o_s, o_e = geno_offsets[o_idx], geno_offsets[o_idx + 1]
            lengths[query, hap] = o_e - o_s
    keep_offsets = np.empty(n_regions * ploidy + 1, np.int64)
    keep_offsets[0] = 0
    keep_offsets[1:] = lengths.cumsum()

    n_variants = keep_offsets[-1]
    groups = np.empty(n_variants, np.uint32)
    group_ends = np.empty(n_variants, np.uint32)
    write_lens = np.empty(n_variants, np.uint32)
    keep = np.empty(n_variants, np.bool_)

    for query in nb.prange(n_regions):
        ref_start: int = starts[query]
        ref_end: int = ends[query]
        for hap in nb.prange(ploidy):
            o_idx = geno_offset_idxs[query, hap]
            o_s, o_e = geno_offsets[o_idx], geno_offsets[o_idx + 1]
            qh_genos = geno_v_idxs[o_s:o_e]
            qh_ccfs = ccfs[o_s:o_e]

            k_idx = query * ploidy + hap
            k_s, k_e = keep_offsets[k_idx], keep_offsets[k_idx + 1]
            qh_groups = groups[k_s:k_e]
            qh_ends = group_ends[k_s:k_e]
            qh_w_lens = write_lens[k_s:k_e]
            qh_keep = keep[k_s:k_e]

            qh_keep[:] = _choose_unphased_variants(
                query_start=ref_start,
                query_end=ref_end,
                variant_idxs=qh_genos,
                ccfs=qh_ccfs,
                positions=positions,
                sizes=sizes,
                groups=qh_groups,
                ref_ends=qh_ends,
                write_lens=qh_w_lens,
                deterministic=deterministic,
            )
    return keep, keep_offsets


@nb.njit(nogil=True, cache=True)
def _choose_unphased_variants(
    query_start: int,
    query_end: int,
    variant_idxs: NDArray[np.int32],  # (v)
    ccfs: NDArray[np.float32],  # (v)
    positions: NDArray[np.int32],  # (total variants)
    sizes: NDArray[np.int32],  # (total variants)
    groups: NDArray[np.uint32],  # (v)
    ref_ends: NDArray[np.uint32],  # (g)
    write_lens: NDArray[np.uint32],  # (g)
    deterministic: bool,
) -> NDArray[np.bool_]:
    # no variants
    if len(variant_idxs) == 0:
        return np.ones(0, np.bool_)

    # treat missing CCF as 1.0
    ccfs = np.nan_to_num(ccfs, True, 1.0)
    groups[:] = UNSEEN_VARIANT
    ref_ends[:] = query_start
    write_lens[:] = 0
    n_groups = 0

    # Assign variants to groups
    # stop once the total written length for all groups is >= length
    for v in range(len(variant_idxs)):
        n_compat = 0
        v_idx: int = variant_idxs[v]
        v_pos = positions[v_idx]
        # +1 assumes atomized variants
        maybe_add_one = int(v_pos >= query_start)
        v_ref_end = v_pos - min(0, sizes[v_idx]) + 1

        if v_ref_end <= query_start:
            # skip the variant by leaving its group as unseen
            continue

        if v_pos >= query_end:
            # variants are sorted by position, everything after this will be outside the query
            break

        # choose group for variant
        for g in range(n_groups):
            # variant compatible with group
            if v_pos >= ref_ends[g]:
                n_compat += 1

                # unseen variant, assign it to first compatible group
                if groups[v] == UNSEEN_VARIANT:
                    groups[v] = g
                # otherwise randomly assign variant to compatible group
                # 1/n_compat chance ensures uniform choice across groups
                elif not deterministic and random.random() < 1 / n_compat:
                    groups[v] = g

        # variant not compatible with any group or there are no groups, make new group
        if groups[v] == UNSEEN_VARIANT:
            n_groups += 1
            groups[v] = n_groups - 1

        # finished with variant assignment

        # update group info
        # writable length = length of variant + dist from last v_ref_end or the missing length, whichever is smaller
        v_group = groups[v]
        # max(v_pos, ref_start) addresses a spanning deletion from before ref_start
        # likewise, (v_pos > ref_start) accounts for a spanning deletion that starts after ref_start
        # spanning deletions will have a write_len of 0 unlike all other variants where we always write at least 1 nt
        # Note that this also assumes no MNPs, only SNPs and INDELs. Can relax this by passing in the alleles similar
        # to above.
        v_write_len = (
            max(v_pos, query_start)
            - ref_ends[v_group]
            + max(0, sizes[v_idx])
            + maybe_add_one
        )
        writable_len = v_write_len
        write_lens[v_group] += writable_len
        ref_ends[v_group] = v_ref_end

    # If not all groups have write_len = target_len after seeing all variants,
    # that's ok since this won't affect the group selection process in the next step.

    if n_groups == 1:
        return groups == 0

    # Choose a group proportional to total ccf normalized by reference length.
    # This is because variants with long ref len will prevent other variants from
    # being assigned to the same group, reducing the potential total ccf of the
    # group.f
    cum_prop = write_lens[:n_groups].view(
        np.float32
    )  # reinterpret this memory to avoid allocation
    for g in range(n_groups):
        v_starts = positions[variant_idxs[groups == g]]
        v_ends = (
            v_starts
            - np.minimum(0, sizes[variant_idxs[groups == g]])
            + (v_starts >= query_start)
        )
        ref_lengths = np.minimum(v_ends, ref_ends[g]) - np.maximum(
            v_starts, query_start
        )
        cum_prop[g] = (ccfs[groups == g] / ref_lengths).sum()
    if deterministic:
        keep_group = cum_prop.argmax()
    else:
        cum_prop = cum_prop.cumsum()
        cum_prop /= cum_prop[-1]
        keep_group = (random.random() <= cum_prop).sum() - 1

    return groups == keep_group



from __future__ import annotations

import re
from functools import partial
from pathlib import Path
from typing import Dict, Iterable, Optional, Union, cast

import numpy as np
import pysam
import seqpro as sp
from numpy.typing import ArrayLike, NDArray
from tqdm.auto import tqdm
from typing_extensions import assert_never

from ._types import Reader
from ._utils import _get_rel_starts, _normalize_contig_name

__all__ = ["Fasta", "NoPadError"]


class NoPadError(Exception):
    pass


class Fasta(Reader):
    dtype = np.dtype("S1")
    """Data type of the sequences."""
    sizes = {}
    """Sizes of non-sequence length dimensions."""
    coords = {}
    """Coordinates of non-sequence length dimensions."""
    chunked = False
    """Whether the reader represents a chunked file format."""

    def __init__(
        self,
        name: str,
        path: Union[str, Path],
        pad: Optional[str] = None,
        alphabet: Optional[Union[str, sp.NucleotideAlphabet, sp.AminoAlphabet]] = None,
        in_memory: bool = False,
        cache: bool = False,
    ) -> None:
        """Read sequences from a FASTA file.

        Parameters
        ----------
        name : str
            Name of the reader, for example `'seq'`.
        path : Union[str, Path]
            Path to the FASTA file.
        pad : Optional[str], optional
            A single character which, if passed, will pad out-of-bound ranges with this
            value. By default no padding is done and out-of-bound ranges raise an error.
        alphabet : str, sp.NucleotideAlphabet, sp.AminoAlphabet, optional
            Alphabet to use for the sequences. If not passed, defaults to DNA.
        in_memory : bool, optional
            Whether to load the sequences into memory. If `True`, the sequences will be
            loaded into memory and the FASTA file will be closed. If `False`, the sequences
            will be read from the FASTA file on demand. Defaults to `False`.
        cache : bool, optional
            Whether to cache the sequences to disk. If `True`, the sequences will be written
            to disk in a `.fa.gvl` file. Defaults to `False`. Only used if `in_memory` is `True`.

        Raises
        ------
        ValueError
            If pad value is not a single character.
        """
        self.name = name
        self.path = Path(path)
        if pad is None:
            self.pad = pad
        else:
            if len(pad) > 1:
                raise ValueError("Pad value must be a single character.")
            self.pad = pad.encode("ascii")

        with self._open() as f:
            self.contigs = {c: f.get_reference_length(c) for c in f.references}

        if alphabet is None:
            self.alphabet: Union[sp.NucleotideAlphabet, sp.AminoAlphabet] = (
                sp.alphabets.DNA
            )
        elif isinstance(alphabet, str):
            alphabet = alphabet.upper()
            try:
                self.alphabet = cast(
                    Union[sp.NucleotideAlphabet, sp.AminoAlphabet],
                    getattr(sp.alphabets, alphabet),
                )
            except AttributeError:
                raise ValueError(f"Alphabet {alphabet} not found.")
        else:
            self.alphabet = alphabet

        if isinstance(self.alphabet, sp.NucleotideAlphabet):
            self.rev_strand_fn = partial(
                self.alphabet.reverse_complement, length_axis=-1
            )
        elif isinstance(self.alphabet, sp.AminoAlphabet):

            def rev_strand_fn(a: NDArray[np.bytes_]):
                return a[::-1]

            self.rev_strand_fn = rev_strand_fn  # type: ignore[assignment]
        else:
            assert_never(self.alphabet)

        self.contigs = self._get_contig_lengths()

        self.handle: Optional[pysam.FastaFile] = None
        fa_extension = re.compile(r"\.(fa|fna|fasta)(\.b?gz)?$")
        self.cache_path = Path(fa_extension.sub(".fa.gvl", str(self.path)))

        if not in_memory:
            self.sequences = None
        else:
            if cache:
                if not self.cache_path.exists():
                    self._write_to_cache()
                self.sequences = self._get_sequences(self.contigs)
            else:
                self.sequences = self._get_sequences(self.contigs)

    def _get_contig_lengths(self) -> Dict[str, int]:
        with self._open() as f:
            return {c: f.get_reference_length(c) for c in f.references}

    def _get_sequences(
        self, contigs: Iterable[str], from_fasta=False
    ) -> Dict[str, NDArray[np.bytes_]]:
        """Load contigs into memory."""
        sequences: Dict[str, NDArray[np.bytes_]] = {}
        if from_fasta or not self.cache_path.exists():
            with self._open() as f:
                pbar = tqdm(total=len(self.contigs))
                for c in contigs:
                    pbar.set_description(f"Reading contig {c}")
                    sequences[c] = np.frombuffer(
                        f.fetch(c).encode("ascii").upper(), "S1"
                    )
                    pbar.update()
                pbar.close()
        elif self.cache_path.exists():
            for contig in contigs:
                sequences[contig] = np.load(self.cache_path / f"{contig}.npy")
        return sequences

    def _write_to_cache(self):
        """Write contigs to cache."""
        self.cache_path.mkdir(parents=True, exist_ok=True)

        with self._open() as f:
            pbar = tqdm(total=len(self.contigs))
            for c in self.contigs:
                if self.sequences is None:
                    pbar.set_description(f"Reading contig {c}")
                    seq = np.frombuffer(f.fetch(c).encode("ascii").upper(), "S1")
                else:
                    seq = self.sequences[c]
                pbar.set_description(f"Writing contig {c}")
                np.save(self.cache_path / f"{c}.npy", seq)
                pbar.update()
            pbar.close()

    def _open(self):
        return pysam.FastaFile(str(self.path))

    def close(self):
        if self.handle is not None:
            self.handle.close()
        self.handle = None

    def read(
        self,
        contig: str,
        starts: ArrayLike,
        ends: ArrayLike,
        **kwargs,
    ) -> NDArray[np.bytes_]:
        """Read a sequence from a FASTA file.

        Parameters
        ----------
        contig : str
            Name of the contig/chromosome.
        starts : ArrayLike
            Start coordinates, 0-based.
        ends : ArrayLike
            End coordinates, 0-based exclusive.
        **kwargs
            Not used.

        Returns
        -------
        NDArray[np.bytes_]
            Shape: (length). Sequence corresponding to the given genomic coordinates.

        Raises
        ------
        ValueError
            Coordinates are out-of-bounds and pad value is not set.
        """
        _contig = _normalize_contig_name(contig, self.contigs)
        if _contig is None:
            raise ValueError(f"Contig {contig} not found.")
        else:
            contig = _contig
        contig_len = self.contigs[contig]

        starts = np.atleast_1d(np.asarray(starts, dtype=np.int64))
        ends = np.atleast_1d(np.asarray(ends, dtype=np.int64))

        if starts.ndim > 1 or ends.ndim > 1:
            raise ValueError("Starts and ends must be coercible to 1D arrays.")

        if len(starts) != len(ends):
            raise ValueError("Starts and ends must be the same length.")

        lengths = ends - starts

        q_starts = np.maximum(0, starts)
        q_ends = np.minimum(self.contigs[contig], ends)

        left_pads = -np.minimum(0, starts)
        right_pads = np.maximum(0, ends - contig_len)

        if self.pad is None:
            if (left_pads > 0).any():
                raise NoPadError("Padding is disabled and a start is < 0.")
            if (right_pads > 0).any():
                raise NoPadError("Padding is disabled and an end is > contig length.")
            out = np.empty(lengths.sum(), dtype="S1")
        else:
            out = np.full(lengths.sum(), self.pad, dtype="S1")

        rel_starts = _get_rel_starts(starts, ends)
        rel_ends = rel_starts + lengths - right_pads
        rel_starts += left_pads

        if self.sequences is None:
            if self.handle is None:
                self.handle = self._open()

            for q_start, q_end, rel_start, rel_end in zip(
                q_starts, q_ends, rel_starts, rel_ends
            ):
                seq = self.handle.fetch(contig, q_start, q_end)
                seq = np.frombuffer(seq.encode("ascii").upper(), "S1")
                out[rel_start:rel_end] = seq
        else:
            for q_start, q_end, rel_start, rel_end in zip(
                q_starts, q_ends, rel_starts, rel_ends
            ):
                seq = self.sequences[contig][q_start:q_end]
                out[rel_start:rel_end] = seq

        return out



from pathlib import Path

import genvarloader as gvl
import numpy as np
from pytest_cases import fixture, parametrize_with_cases

data_dir = Path(__file__).resolve().parents[1] / "data"
ref = data_dir / "fasta" / "Homo_sapiens.GRCh38.dna.primary_assembly.fa.bgz"

def ds_phased():
    return gvl.Dataset.open(data_dir / "phased_dataset.gvl", ref)


def ds_unphased():
    return gvl.Dataset.open(data_dir / "unphased_dataset.gvl", ref)


def seqs_ref():
    return "reference"


def seqs_haps():
    return "haplotypes"


def seqs_annot():
    return "annotated"


def bool_false():
    return False


def bool_true():
    return True


@fixture(scope="session")
@parametrize_with_cases("ds", prefix="ds_", cases=".")
@parametrize_with_cases("seq_type", prefix="seqs_", cases=".")
@parametrize_with_cases("return_indices", prefix="bool_", cases=".")
def dataset(ds: gvl.Dataset, seq_type, return_indices: bool):
    return ds.with_seqs(seq_type).with_indices(return_indices)


def idx_scalar():
    return 0


def idx_neg_scalar():
    return -1


def idx_slice_none():
    return slice(None)


def idx_slice_start_none():
    return slice(1, None)


def idx_slice_none_stop():
    return slice(None, 2)


def idx_list():
    return [0, 1, 2]


def idx_array():
    return np.arange(3)


@parametrize_with_cases("idx", prefix="idx_", cases=".")
def test_ds_indexing(dataset, idx):
    dataset[idx]


@parametrize_with_cases("r_idx", prefix="idx_", cases=".")
@parametrize_with_cases("s_idx", prefix="idx_", cases=".")
def test_rs_indexing(dataset, r_idx, s_idx):
    dataset[r_idx, s_idx]
